import asyncio
import os
import pathlib
import secrets
import signal
import threading

from functools import partial
from multiprocessing import Process
from time import time
from typing import Mapping, Optional

import psutil

from tornado.httpserver import HTTPServer
from tornado.netutil import bind_sockets
from tornado.process import cpu_count, fork_processes
from tornado.wsgi import WSGIContainer
from tornado.web import Application, FallbackHandler

from platypush.backend import Backend
from platypush.backend.http.app import application
from platypush.backend.http.app.utils import get_streaming_routes, get_ws_routes
from platypush.backend.http.app.ws.events import WSEventProxy
from platypush.bus.redis import RedisBus
from platypush.config import Config
from platypush.utils import get_remaining_timeout


class HttpBackend(Backend):
    """
    The HTTP backend is a general-purpose web server.

    Example configuration:

      .. code-block:: yaml

        backend.http:
            # Default HTTP listen port
            port: 8008
            # External folders that will be exposed over `/resources/<name>`
            resource_dirs:
                photos: /mnt/hd/photos
                videos: /mnt/hd/videos
                music: /mnt/hd/music

    You can leverage this backend:

        * To execute Platypush commands via HTTP calls. In order to do so:

            * Register a user to Platypush through the web panel (usually
              served on ``http://host:8008/``).

            * Generate a token for your user, either through the web panel
              (Settings -> Generate Token) or via API:

                .. code-block:: shell

                    curl -XPOST -H 'Content-Type: application/json' -d '
                      {
                        "username": "$YOUR_USER",
                        "password": "$YOUR_PASSWORD"
                      }' http://host:8008/auth

            * Execute actions through the ``/execute`` endpoint:

                .. code-block:: shell

                    curl -XPOST -H 'Content-Type: application/json' -H "Authorization: Bearer $YOUR_TOKEN" -d '
                      {
                        "type": "request",
                        "action": "tts.say",
                        "args": {
                          "text": "This is a test"
                        }
                      }' http://host:8008/execute

        * To interact with your system (and control plugins and backends)
          through the Platypush web panel, by default available on
          ``http://host:8008/``. Any configured plugin that has an available
          panel plugin will be automatically added to the web panel.

        * To create asynchronous integrations with Platypush over websockets.
          Two routes are available:

            * ``/ws/events`` - Subscribe to this websocket to receive the
              events generated by the application.
            * ``/ws/requests`` - Subscribe to this websocket to send commands
              to Platypush and receive the response asynchronously.

          You will have to authenticate your connection to these websockets,
          just like the ``/execute`` endpoint. In both cases, you can pass the
          token either via ``Authorization: Bearer``, via the ``token`` query
          string or body parameter, or leverage ``Authorization: Basic`` with
          username and password (not advised), or use a valid ``session_token``
          cookie from an authenticated web panel session.

        * To display a fullscreen dashboard with custom widgets.

            * Widgets are available as Vue.js components under
              ``platypush/backend/http/webapp/src/components/widgets``.

            * Explore their options (some may require some plugins or backends
              to be configured in order to work) and create a new dashboard
              template under ``~/.config/platypush/dashboards``- e.g.
              ``main.xml``:

                .. code-block:: xml

                    <Dashboard>
                        <!-- Display the following widgets on the same row. Each row consists of 12 columns.
                             You can specify the width of each widget either through class name (e.g. col-6 means
                             6 columns out of 12, e.g. half the size of the row) or inline style
                             (e.g. `style="width: 50%"`). -->
                        <Row>
                            <!-- Show a calendar widget with the upcoming events. It requires the `calendar` plugin to
                                 be enabled and configured. -->
                            <Calendar class="col-6" />

                            <!-- Show the current track and other playback info. It requires `music.mpd` plugin or any
                                 other music plugin enabled. -->
                            <Music class="col-3" />

                            <!-- Show current date, time and weather.
                                 It requires a `weather` plugin or backend enabled -->
                            <DateTimeWeather class="col-3" />
                        </Row>

                        <!-- Display the following widgets on a second row -->
                        <Row>
                            <!-- Show a carousel of images from a local folder. For security reasons, the folder must be
                                 explicitly exposed as an HTTP resource through the backend
                                 `resource_dirs` attribute. -->
                            <ImageCarousel class="col-6" img-dir="/mnt/hd/photos/carousel" />

                            <!-- Show the news headlines parsed from a list of RSS feed and stored locally through the
                                 `http.poll` backend -->
                            <RssNews class="col-6" db="sqlite:////path/to/your/rss.db" />
                        </Row>
                    </Dashboard>

            * The dashboard will be accessible under
              ``http://host:8008/dashboard/<name>``, where ``name=main`` if for
              example you stored your template under
              ``~/.config/platypush/dashboards/main.xml``.

        * To expose custom endpoints that can be called as web hooks by other
          applications and run some custom logic. All you have to do in this case
          is to create a hook on a
          :class:`platypush.message.event.http.hook.WebhookEvent` with the
          endpoint that you want to expose and store it under e.g.
          ``~/.config/platypush/scripts/hooks.py``:

            .. code-block:: python

                from platypush import get_plugin, when
                from platypush.message.event.http.hook import WebhookEvent

                hook_token = 'abcdefabcdef'

                # Expose the hook under the /hook/lights_toggle endpoint
                @when(WebhookEvent, hook='lights_toggle')
                def lights_toggle(event, **context):
                    # Do any checks on the request
                    assert event.headers.get('X-Token') == hook_token, 'Unauthorized'

                    # Run some actions
                    lights = get_plugin('light.hue')
                    lights.toggle()

    Any plugin can register custom routes under ``platypush/backend/http/app/routes/plugins``.
    Any additional route is managed as a Flask blueprint template and the `.py`
    module can expose lists of routes to the main webapp through the
    ``__routes__`` object (a list of Flask blueprints).

    Security: Access to the endpoints requires at least one user to be
    registered. Access to the endpoints is regulated in the following ways
    (with the exception of event hooks, whose logic is up to the user):

        * **Simple authentication** - i.e. registered username and password.
        * **JWT token** provided either over as ``Authorization: Bearer``
          header or ``GET`` ``?token=<TOKEN>`` parameter. A JWT token can be
          generated either through the web panel or over the ``/auth``
          endpoint.
        * **Global platform token**, usually configured on the root of the
          ``config.yaml`` as ``token: <VALUE>``. It can provided either over on
          the ``X-Token`` header or as a ``GET`` ``?token=<TOKEN>`` parameter.
        * **Session token**, generated upon login, it can be used to
          authenticate requests through the ``Cookie`` header (cookie name:
          ``session_token``).

    """

    DEFAULT_HTTP_PORT = 8008
    """The default listen port for the webserver."""

    _STOP_TIMEOUT = 5
    """How long we should wait (in seconds) before killing the worker processes."""

    def __init__(
        self,
        port: int = DEFAULT_HTTP_PORT,
        bind_address: str = '0.0.0.0',
        resource_dirs: Optional[Mapping[str, str]] = None,
        secret_key_file: Optional[str] = None,
        num_workers: Optional[int] = None,
        use_werkzeug_server: bool = False,
        **kwargs,
    ):
        """
        :param port: Listen port for the web server (default: 8008)
        :param bind_address: Address/interface to bind to (default: 0.0.0.0, accept connection from any IP)
        :param resource_dirs: Static resources directories that will be
            accessible through ``/resources/<path>``. It is expressed as a map
            where the key is the relative path under ``/resources`` to expose and
            the value is the absolute path to expose.
        :param secret_key_file: Path to the file containing the secret key that will be used by Flask
            (default: ``~/.local/share/platypush/flask.secret.key``).
        :param num_workers: Number of worker processes to use (default: ``(cpu_count * 2) + 1``).
        :param use_werkzeug_server: Whether the backend should be served by a
            Werkzeug server (default: ``False``). Note that using the built-in
            Werkzeug server instead of Tornado is very inefficient, and it
            doesn't support websocket-based features either so the UI will
            probably be severely limited. You should only use this option if:

                - You are running tests.
                - You have issues with running a full Tornado server - for
                  example, you are running the application on a small embedded
                  device that doesn't support Tornado.
        """

        super().__init__(**kwargs)

        self.port = port
        self._server_proc: Optional[Process] = None
        self._service_registry_thread = None
        self.bind_address = bind_address

        if resource_dirs:
            self.resource_dirs = {
                name: os.path.abspath(os.path.expanduser(d))
                for name, d in resource_dirs.items()
            }
        else:
            self.resource_dirs = {}

        self.secret_key_file = os.path.expanduser(
            secret_key_file
            or os.path.join(Config.get('workdir'), 'flask.secret.key')  # type: ignore
        )
        self.local_base_url = f'http://localhost:{self.port}'
        self.num_workers = num_workers or (cpu_count() * 2) + 1
        self.use_werkzeug_server = use_werkzeug_server

    def send_message(self, *_, **__):
        self.logger.warning('Use cURL or any HTTP client to query the HTTP backend')

    def on_stop(self):
        """On backend stop"""
        super().on_stop()
        self.logger.info('Received STOP event on HttpBackend')
        start = time()
        remaining_time: partial[float] = partial(  # type: ignore
            get_remaining_timeout, timeout=self._STOP_TIMEOUT, start=start
        )

        if self._server_proc:
            if self._server_proc.pid:
                try:
                    os.kill(self._server_proc.pid, signal.SIGINT)
                except OSError:
                    pass

            if self._server_proc and self._server_proc.is_alive():
                self._server_proc.join(timeout=remaining_time() / 2)
                try:
                    self._server_proc.terminate()
                    self._server_proc.join(timeout=remaining_time() / 2)
                except AttributeError:
                    pass

        if self._server_proc and self._server_proc.is_alive():
            self._server_proc.kill()

        self._server_proc = None

        if self._service_registry_thread and self._service_registry_thread.is_alive():
            self._service_registry_thread.join(timeout=remaining_time())
            self._service_registry_thread = None

        self.logger.info('HTTP server terminated')

    def notify_web_clients(self, event):
        """Notify all the connected web clients (over websocket) of a new event"""
        WSEventProxy.publish(event)  # noqa: E1120

    def _get_secret_key(self, _create=False):
        if _create:
            self.logger.info('Creating web server secret key')
            pathlib.Path(self.secret_key_file).parent.mkdir(parents=True, exist_ok=True)
            with open(self.secret_key_file, 'w') as f:
                f.write(secrets.token_urlsafe(32))

            os.chmod(self.secret_key_file, 0o600)
            return secrets.token_urlsafe(32)

        try:
            with open(self.secret_key_file, 'r') as f:
                return f.read()
        except IOError as e:
            if not _create:
                return self._get_secret_key(_create=True)

            raise e

    def _register_service(self):
        try:
            self.register_service(port=self.port)
        except Exception as e:
            self.logger.warning('Could not register the Zeroconf service')
            self.logger.exception(e)

    def _start_zeroconf_service(self):
        self._service_registry_thread = threading.Thread(
            target=self._register_service,
            name='ZeroconfService',
        )
        self._service_registry_thread.start()

    async def _post_fork_main(self, sockets):
        assert isinstance(
            self.bus, RedisBus
        ), 'The HTTP backend only works if backed by a Redis bus'

        application.config['redis_queue'] = self.bus.redis_queue
        application.secret_key = self._get_secret_key()
        container = WSGIContainer(application)
        tornado_app = Application(
            [
                *[
                    (route.path(), route)
                    for route in [*get_ws_routes(), *get_streaming_routes()]
                ],
                (r'.*', FallbackHandler, {'fallback': container}),
            ]
        )

        server = HTTPServer(tornado_app)
        server.add_sockets(sockets)

        try:
            await asyncio.Event().wait()
        except (asyncio.CancelledError, KeyboardInterrupt):
            pass
        finally:
            server.stop()
            await server.close_all_connections()

    def _web_server_proc(self):
        self.logger.info(
            'Starting local web server on port %s with %d service workers',
            self.port,
            self.num_workers,
        )

        if self.use_werkzeug_server:
            application.config['redis_queue'] = self.bus.redis_queue  # type: ignore
            application.run(
                host=self.bind_address,
                port=self.port,
                use_reloader=False,
                debug=True,
            )
        else:
            sockets = bind_sockets(
                self.port, address=self.bind_address, reuse_port=True
            )

            try:
                fork_processes(self.num_workers)
                future = self._post_fork_main(sockets)
                asyncio.run(future)
            except (asyncio.CancelledError, KeyboardInterrupt):
                pass
            finally:
                self._stop_workers()

    def _stop_workers(self):
        """
        Stop all the worker processes.

        We have to run this manually on server termination because of a
        long-standing issue with Tornado not being able to wind down the forked
        workers when the server terminates:
        https://github.com/tornadoweb/tornado/issues/1912.
        """
        parent_pid = (
            self._server_proc.pid
            if self._server_proc and self._server_proc.pid
            else None
        )

        if not parent_pid:
            return

        try:
            cur_proc = psutil.Process(parent_pid)
        except psutil.NoSuchProcess:
            return

        # Send a SIGTERM to all the children
        children = cur_proc.children()
        for child in children:
            if child.pid != parent_pid and child.is_running():
                try:
                    os.kill(child.pid, signal.SIGTERM)
                except OSError as e:
                    self.logger.warning(
                        'Could not send SIGTERM to PID %d: %s', child.pid, e
                    )

        # Initialize the timeout
        start = time()
        remaining_time: partial[int] = partial(  # type: ignore
            get_remaining_timeout, timeout=self._STOP_TIMEOUT, start=start, cls=int
        )

        # Wait for all children to terminate (with timeout)
        for child in children:
            if child.pid != parent_pid and child.is_running():
                try:
                    child.wait(timeout=remaining_time())
                except TimeoutError:
                    pass

        # Send a SIGKILL to any child process that is still running
        for child in children:
            if child.pid != parent_pid and child.is_running():
                try:
                    child.kill()
                except OSError:
                    pass

    def _start_web_server(self):
        self._server_proc = Process(target=self._web_server_proc)
        self._server_proc.start()
        self._server_proc.join()

    def run(self):
        super().run()

        self._start_zeroconf_service()
        self._start_web_server()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
import logging
from multiprocessing import Event, Process, RLock
from os import getpid
from typing import Optional


class ControllableProcess(Process, ABC):
    """
    Extends the ``Process`` class to allow for external extended control of the
    underlying process.
    """

    _kill_timeout: float = 10.0

    def __init__(self, *args, timeout: Optional[float] = None, **kwargs):
        kwargs['name'] = kwargs.get('name', self.__class__.__name__)
        super().__init__(*args, **kwargs)

        self.timeout = timeout
        self.logger = logging.getLogger(self.name)
        self._should_stop = Event()
        self._stop_lock = RLock()
        self._should_restart = False

    @property
    def _timeout_err(self) -> TimeoutError:
        return TimeoutError(f'Process {self.name} timed out')

    @property
    def should_stop(self) -> bool:
        """
        :return: ``True`` if the process is scheduled for stop, ``False``
            otherwise.
        """
        return self._should_stop.is_set()

    def wait_stop(self, timeout: Optional[float] = None) -> None:
        """
        Waits for the process to stop.

        :param timeout: The maximum time to wait for the process to stop.
        """
        timeout = timeout if timeout is not None else self.timeout
        stopped = self._should_stop.wait(timeout=timeout)

        if not stopped:
            raise self._timeout_err

        if self.pid == getpid():
            return  # Prevent termination deadlock

        self.join(timeout=timeout)
        if self.is_alive():
            raise self._timeout_err

    def stop(self, timeout: Optional[float] = None) -> None:
        """
        Stops the process.

        :param timeout: The maximum time to wait for the process to stop.
        """
        timeout = timeout if timeout is not None else self._kill_timeout
        with self._stop_lock:
            self._should_stop.set()
            self.on_stop()

            try:
                if self.pid != getpid():
                    self.wait_stop(timeout=timeout)
            except TimeoutError:
                pass
            finally:
                self.terminate()

            try:
                if self.pid != getpid():
                    self.wait_stop(timeout=self._kill_timeout)
            except TimeoutError:
                self.logger.warning(
                    'The process %s is still alive after %f seconds, killing it',
                    self.name,
                    self._kill_timeout,
                )
                self.kill()

    def on_stop(self) -> None:
        """
        Handler called when the process is stopped.

        It can be implemented by subclasses.
        """

    @property
    def should_restart(self) -> bool:
        """
        :return: ``True`` if the process is marked for restart after stop,
            ``False`` otherwise.
        """
        return self._should_restart

    def mark_for_restart(self):
        """
        Marks the process for restart after stop.
        """
        self._should_restart = True

    @abstractmethod
    def main(self):
        """
        The main function of the process.

        It must be implemented by subclasses.
        """

    def _main(self):
        """
        Wrapper for the main function of the process.
        """
        self._should_restart = False
        return self.main()

    def run(self) -> None:
        """
        Executes the process.
        """
        super().run()

        try:
            self._main()
        finally:
            self._should_stop.set()
            self.logger.info('Process terminated')



from platypush.context import get_plugin
from platypush.plugins.media.search import MediaSearcher


# pylint: disable=too-few-public-methods
class YoutubeMediaSearcher(MediaSearcher):
    """
    Search YouTube videos by query.
    """

    def search(self, query: str, *_, **__):
        """
        Performs a YouTube search using the ``youtube`` plugin.
        """

        self.logger.info('Searching YouTube for "%s"', query)
        yt = get_plugin('youtube')
        assert yt, 'YouTube plugin not available/configured'
        return yt.search(query=query).output


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from platypush.schemas.dataclasses import percent_field


@dataclass
class CpuInfo:
    """
    CPU info data class.
    """

    architecture: Optional[str] = field(
        metadata={
            'data_key': 'arch_string_raw',
            'metadata': {
                'description': 'CPU architecture',
                'example': 'x86_64',
            },
        }
    )

    bits: int = field(
        metadata={
            'metadata': {
                'description': 'CPU bits / register size',
                'example': 64,
            }
        }
    )

    cores: int = field(
        metadata={
            'data_key': 'count',
            'metadata': {
                'description': 'Number of cores',
                'example': 4,
            },
        }
    )

    vendor: Optional[str] = field(
        metadata={
            'data_key': 'vendor_id_raw',
            'metadata': {
                'description': 'Vendor string',
                'example': 'GenuineIntel',
            },
        }
    )

    brand: Optional[str] = field(
        metadata={
            'data_key': 'brand_raw',
            'metadata': {
                'description': 'CPU brand string',
                'example': 'Intel(R) Core(TM) i7-5500U CPU @ 2.40GHz',
            },
        }
    )

    frequency_advertised: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Advertised CPU frequency, in Hz',
                'example': 2400000000,
            }
        }
    )

    frequency_actual: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Actual CPU frequency, in Hz',
                'example': 2350000000,
            }
        }
    )

    flags: List[str] = field(
        metadata={
            'metadata': {
                'description': 'CPU flags',
                'example': ['acpi', 'aes', 'cpuid'],
            }
        }
    )

    l1_instruction_cache_size: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Size of the L1 instruction cache, in bytes',
                'example': 65536,
            }
        }
    )

    l1_data_cache_size: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Size of the L1 data cache, in bytes',
                'example': 65536,
            }
        }
    )

    l2_cache_size: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Size of the L2 cache, in bytes',
                'example': 524288,
            }
        }
    )

    l3_cache_size: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Size of the L2 cache, in bytes',
                'example': 4194304,
            }
        }
    )


@dataclass
class CpuTimes:
    """
    CPU times data class.
    """

    user: Optional[float] = percent_field()
    nice: Optional[float] = percent_field()
    system: Optional[float] = percent_field()
    idle: Optional[float] = percent_field()
    iowait: Optional[float] = percent_field()
    irq: Optional[float] = percent_field()
    softirq: Optional[float] = percent_field()
    steal: Optional[float] = percent_field()
    guest: Optional[float] = percent_field()
    guest_nice: Optional[float] = percent_field()


@dataclass
class CpuStats:
    """
    CPU stats data class.
    """

    ctx_switches: int
    interrupts: int
    soft_interrupts: int
    syscalls: int


@dataclass
class CpuFrequency:
    """
    CPU frequency data class.
    """

    current: float
    min: float
    max: float


@dataclass
class Cpu:
    """
    CPU data aggregate dataclass.
    """

    info: CpuInfo
    times: CpuTimes
    frequency: CpuFrequency
    stats: CpuStats
    load_avg: Tuple[float, float, float]
    percent: float = percent_field()



import logging
import threading

from platypush.bus import Bus
from platypush.message import Message

logger = logging.getLogger('platypush:bus:redis')


class RedisBus(Bus):
    """
    Overrides the in-process in-memory local bus with a Redis bus
    """

    DEFAULT_REDIS_QUEUE: str = 'platypush/bus'

    def __init__(self, *args, on_message=None, redis_queue=None, **kwargs):
        from platypush.utils import get_redis

        super().__init__(on_message=on_message)
        self.redis = get_redis(*args, **kwargs)
        self.redis_args = kwargs
        self.redis_queue = redis_queue or self.DEFAULT_REDIS_QUEUE
        self.on_message = on_message
        self.thread_id = threading.get_ident()
        self._pubsub = None
        self._pubsub_lock = threading.RLock()

    @property
    def pubsub(self):
        with self._pubsub_lock:
            if not self._pubsub:
                self._pubsub = self.redis.pubsub()
            return self._pubsub

    def poll(self):
        """
        Polls the Redis queue for new messages
        """
        with self.pubsub as pubsub:
            pubsub.subscribe(self.redis_queue)
            try:
                for msg in pubsub.listen():
                    if msg.get('type') != 'message':
                        continue

                    if self.should_stop():
                        break

                    try:
                        data = msg.get('data', b'').decode('utf-8')
                        parsed_msg = Message.build(data)
                        if parsed_msg and self.on_message:
                            self.on_message(parsed_msg)
                    except Exception as e:
                        logger.exception(e)
            finally:
                pubsub.unsubscribe(self.redis_queue)

    def post(self, msg):
        """
        Sends a message to the Redis queue
        """
        from redis import exceptions

        try:
            self.redis.publish(self.redis_queue, str(msg))
        except exceptions.ConnectionError as e:
            if not self.should_stop():
                # Raise the exception only if the bus it not supposed to be
                # stopped
                raise e

    def stop(self):
        super().stop()
        self.redis.close()


# vim:sw=4:ts=4:et:



from ._errors import Errors
from ._invite import Invite, RoomInvite, UserInvite
from ._presence import XmppPresence


__all__ = [
    "Errors",
    "Invite",
    "RoomInvite",
    "UserInvite",
    "XmppPresence",
]



from sqlalchemy import JSON, Column, DateTime, Float, Integer, ForeignKey, String

from . import Entity


class Weather(Entity):
    """
    Weather entity.
    """

    __tablename__ = 'weather'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    summary = Column(String)
    icon = Column(String)
    image = Column(String)
    precip_intensity = Column(Float)
    precip_type = Column(String)
    temperature = Column(Float)
    apparent_temperature = Column(Float)
    humidity = Column(Float)
    pressure = Column(Float)
    rain_chance = Column(Float)
    wind_speed = Column(Float)
    wind_direction = Column(String)
    wind_gust = Column(Float)
    cloud_cover = Column(Float)
    visibility = Column(Float)
    sunrise = Column(DateTime)
    sunset = Column(DateTime)
    units = Column(String)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class WeatherForecast(Entity):
    """
    Weather forecast entity.
    """

    __tablename__ = 'weather_forecast'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)
    # forecast contains a list of serialized Weather entities
    forecast = Column(JSON)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }



import queue
import pytest
import time

test_timeout = 10
cron_queue = queue.Queue()


def _test_cron_queue(expected_msg: str):
    msg = None
    test_start = time.time()
    while time.time() - test_start <= test_timeout and msg != expected_msg:
        try:
            msg = cron_queue.get(block=True, timeout=test_timeout)
        except queue.Empty:
            break

    assert msg == expected_msg, 'The expected cronjob has not been executed'


def test_cron_execution():
    """
    Test that the cronjob in ``../etc/scripts/test_cron.py`` runs successfully.
    """
    _test_cron_queue('cron_test')


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



from typing import List

from platypush.config import Config
from platypush.context import get_plugin
from platypush.plugins.media import MediaPlugin


audio_plugins = [
    'sound',
    'music.mpd',
]


def get_default_media_plugin(video: bool = False) -> MediaPlugin:
    """
    Get the default media plugin based on the current configuration.

    :param video: If True then the plugin must support video playback.
    """

    enabled_plugins: List[MediaPlugin] = []
    cfg = Config.get() or {}

    for plugin_name in MediaPlugin.supported_media_plugins:
        try:
            plugin = get_plugin(plugin_name)
            if plugin and plugin_name in cfg and not cfg[plugin_name].get('disabled'):
                enabled_plugins.append(plugin)
        except Exception:
            pass

    local_plugins = [plugin for plugin in enabled_plugins if plugin.is_local()]

    if local_plugins:
        return local_plugins[0]

    if enabled_plugins:
        return enabled_plugins[0]

    assert not video, (
        'No media plugin with video support is enabled. '
        f'Supported plugins: {MediaPlugin.supported_media_plugins}'
    )

    for plugin_name in audio_plugins:
        try:
            plugin = get_plugin(plugin_name)
            if plugin and plugin_name in cfg and not cfg[plugin_name].get('disabled'):
                return plugin
        except Exception:
            pass

    raise AssertionError('No media plugin is enabled')



import inspect
import pathlib
import subprocess
from functools import lru_cache
from typing import List, Optional

from platypush.commands import CommandStream, RestartCommand, StopCommand
from platypush.common.db import override_definitions
from platypush.config import Config
from platypush.plugins import Plugin, action
from platypush.utils import get_backend_class_by_name, get_plugin_class_by_name
from platypush.utils.manifest import Manifest
from platypush.utils.mock import auto_mocks


class ApplicationPlugin(Plugin):
    """
    This plugin is used to control and inspect the application state.
    """

    @property
    def _ctrl_sock(self) -> Optional[str]:
        """
        :return: The path to the UNIX socket to control the application.
        """
        return Config.get('ctrl_sock')  # type: ignore

    @action
    def stop(self):
        """
        Stop the application.
        """
        CommandStream(self._ctrl_sock).write(StopCommand())

    @action
    def restart(self):
        """
        Restart the application.
        """
        CommandStream(self._ctrl_sock).write(RestartCommand())

    @action
    def install(self, extension: str):
        """
        Install the dependencies of an extension.

        :param extension: Extension name. For plugins, it will be the plugin
            name (e.g. ``light.hue`` or ``music.mpd``); for backend, the name
            will be prefixed by ``backend.`` (e.g. ``backend.http`` or
            ``backend.tcp``).
        """
        install_cmds = self._get_install_cmds(extension)
        if not install_cmds:
            self.logger.info('No extra requirements found for extension %s', extension)
            return

        for cmd in install_cmds:
            self.logger.info('> %s', cmd)
            subprocess.check_call(cmd, shell=True, stderr=subprocess.STDOUT)

    @action
    def get_install_commands(self, extension: str) -> List[str]:
        """
        Get the installation commands for an extension.

        :param extension: Extension name. For plugins, it will be the plugin
            name (e.g. ``light.hue`` or ``music.mpd``); for backend, the name
            will be prefixed by ``backend.`` (e.g. ``backend.http`` or
            ``backend.tcp``).
        """
        return self._get_install_cmds(extension)

    @lru_cache(maxsize=256)  # noqa
    def _get_install_cmds(self, extension: str) -> List[str]:
        getter = get_plugin_class_by_name
        if extension.startswith('backend.'):
            extension = extension[len('backend.') :]
            getter = get_backend_class_by_name

        with auto_mocks(), override_definitions():
            ext = getter(extension)

        assert ext, f'Could not find extension {extension}'
        manifest_file = str(pathlib.Path(inspect.getfile(ext)).parent / 'manifest.json')
        return list(Manifest.from_file(manifest_file).install.to_install_commands())



from marshmallow import fields
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class MillisecondsTimestamp(DateTime):
    def _get_attr(self, *args, **kwargs):
        value = super()._get_attr(*args, **kwargs)
        if isinstance(value, int):
            value = float(value / 1000)
        return value


class MatrixEventIdSchema(Schema):
    event_id = fields.String(
        required=True,
        metadata={
            'description': 'Event ID',
            'example': '$24KT_aQz6sSKaZH8oTCibRTl62qywDgQXMpz5epXsW5',
        },
    )


class MatrixRoomIdSchema(Schema):
    room_id = fields.String(
        required=True,
        metadata={
            'description': 'Room ID',
            'example': '!aBcDeFgHiJkMnO:matrix.example.org',
        },
    )


class MatrixProfileSchema(Schema):
    user_id = fields.String(
        required=True,
        metadata={
            'description': 'User ID',
            'example': '@myuser:matrix.example.org',
        },
    )

    display_name = fields.String(
        attribute='displayname',
        metadata={
            'description': 'User display name',
            'example': 'Foo Bar',
        },
    )

    avatar_url = fields.URL(
        metadata={
            'description': 'User avatar URL',
            'example': 'mxc://matrix.example.org/AbCdEfG0123456789',
        }
    )


class MatrixMemberSchema(MatrixProfileSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['display_name'].attribute = 'display_name'


class MatrixRoomSchema(Schema):
    room_id = fields.String(
        required=True,
        metadata={
            'description': 'Room ID',
            'example': '!aBcDeFgHiJkMnO:matrix.example.org',
        },
    )

    name = fields.String(
        metadata={
            'description': 'Room name',
            'example': 'My Room',
        }
    )

    display_name = fields.String(
        metadata={
            'description': 'Room display name',
            'example': 'My Room',
        }
    )

    topic = fields.String(
        metadata={
            'description': 'Room topic',
            'example': 'My Room Topic',
        }
    )

    avatar_url = fields.URL(
        attribute='room_avatar_url',
        metadata={
            'description': 'Room avatar URL',
            'example': 'mxc://matrix.example.org/AbCdEfG0123456789',
        },
    )

    owner_id = fields.String(
        attribute='own_user_id',
        metadata={
            'description': 'Owner user ID',
            'example': '@myuser:matrix.example.org',
        },
    )

    encrypted = fields.Bool()


class MatrixDeviceSchema(Schema):
    device_id = fields.String(
        required=True,
        attribute='id',
        metadata={
            'description': 'ABCDEFG',
        },
    )

    user_id = fields.String(
        required=True,
        metadata={
            'description': 'User ID associated to the device',
            'example': '@myuser:matrix.example.org',
        },
    )

    display_name = fields.String(
        metadata={
            'description': 'Display name of the device',
            'example': 'Element Android',
        },
    )

    blacklisted = fields.Boolean()
    deleted = fields.Boolean(default=False)
    ignored = fields.Boolean()
    verified = fields.Boolean()

    keys = fields.Dict(
        metadata={
            'description': 'Encryption keys supported by the device',
            'example': {
                'curve25519': 'BtlB0vaQmtYFsvOYkmxyzw9qP5yGjuAyRh4gXh3q',
                'ed25519': 'atohIK2FeVlYoY8xxpZ1bhDbveD+HA2DswNFqUxP',
            },
        },
    )


class MatrixMyDeviceSchema(Schema):
    device_id = fields.String(
        required=True,
        attribute='id',
        metadata={
            'description': 'ABCDEFG',
        },
    )

    display_name = fields.String(
        metadata={
            'description': 'Device display name',
            'example': 'My Device',
        }
    )

    last_seen_ip = fields.String(
        metadata={
            'description': 'Last IP associated to this device',
            'example': '1.2.3.4',
        }
    )

    last_seen_date = DateTime(
        metadata={
            'description': 'The last time that the device was reported online',
            'example': '2022-07-23T17:20:01.254223',
        }
    )


class MatrixDownloadedFileSchema(Schema):
    url = fields.String(
        metadata={
            'description': 'Matrix URL of the original resource',
            'example': 'mxc://matrix.example.org/YhQycHvFOvtiDDbEeWWtEhXx',
        },
    )

    path = fields.String(
        metadata={
            'description': 'Local path where the file has been saved',
            'example': '/home/user/Downloads/image.png',
        }
    )

    content_type = fields.String(
        metadata={
            'description': 'Content type of the downloaded file',
            'example': 'image/png',
        }
    )

    size = fields.Int(
        metadata={
            'description': 'Length in bytes of the output file',
            'example': 1024,
        }
    )


class MatrixMessageSchema(Schema):
    event_id = fields.String(
        required=True,
        metadata={
            'description': 'Event ID associated to this message',
            'example': '$2eOQ5ueafANj91GnPCRkRUOOjM7dI5kFDOlfMNCD2ly',
        },
    )

    room_id = fields.String(
        required=True,
        metadata={
            'description': 'The ID of the room containing the message',
            'example': '!aBcDeFgHiJkMnO:matrix.example.org',
        },
    )

    user_id = fields.String(
        required=True,
        attribute='sender',
        metadata={
            'description': 'ID of the user who sent the message',
            'example': '@myuser:matrix.example.org',
        },
    )

    body = fields.String(
        required=True,
        metadata={
            'description': 'Message body',
            'example': 'Hello world!',
        },
    )

    format = fields.String(
        metadata={
            'description': 'Message format',
            'example': 'markdown',
        },
    )

    formatted_body = fields.String(
        metadata={
            'description': 'Formatted body',
            'example': '**Hello world!**',
        },
    )

    url = fields.String(
        metadata={
            'description': 'mxc:// URL if this message contains an attachment',
            'example': 'mxc://matrix.example.org/oarGdlpvcwppARPjzNlmlXkD',
        },
    )

    content_type = fields.String(
        attribute='mimetype',
        metadata={
            'description': 'If the message contains an attachment, this field '
            'will contain its MIME type',
            'example': 'image/jpeg',
        },
    )

    transaction_id = fields.String(
        metadata={
            'description': 'Set if this message a unique transaction_id associated',
            'example': 'mQ8hZR6Dx8I8YDMwONYmBkf7lTgJSMV/ZPqosDNM',
        },
    )

    decrypted = fields.Bool(
        metadata={
            'description': 'True if the message was encrypted and has been '
            'successfully decrypted',
        },
    )

    verified = fields.Bool(
        metadata={
            'description': 'True if this is an encrypted message coming from a '
            'verified source'
        },
    )

    hashes = fields.Dict(
        metadata={
            'description': 'If the message has been decrypted, this field '
            'contains a mapping of its hashes',
            'example': {'sha256': 'yoQLQwcURq6/bJp1xQ/uhn9Z2xeA27KhMhPd/mfT8tR'},
        },
    )

    iv = fields.String(
        metadata={
            'description': 'If the message has been decrypted, this field '
            'contains the encryption initial value',
            'example': 'NqJMMdijlLvAAAAAAAAAAA',
        },
    )

    key = fields.Dict(
        metadata={
            'description': 'If the message has been decrypted, this field '
            'contains the encryption/decryption key',
            'example': {
                'alg': 'A256CTR',
                'ext': True,
                'k': 'u6jjAyNvJoBHE55P5ZfvX49m3oSt9s_L4PSQdprRSJI',
                'key_ops': ['encrypt', 'decrypt'],
                'kty': 'oct',
            },
        },
    )

    timestamp = MillisecondsTimestamp(
        required=True,
        attribute='server_timestamp',
        metadata={
            'description': 'When the event was registered on the server',
            'example': '2022-07-23T17:20:01.254223',
        },
    )


class MatrixMessagesResponseSchema(Schema):
    messages = fields.Nested(
        MatrixMessageSchema,
        many=True,
        required=True,
        attribute='chunk',
    )

    start = fields.String(
        required=True,
        nullable=True,
        metadata={
            'description': 'Pointer to the first message. It can be used as a '
            '``start``/``end`` for another ``get_messages`` query.',
            'example': 's10226_143893_619_3648_5951_5_555_7501_0',
        },
    )

    end = fields.String(
        required=True,
        nullable=True,
        metadata={
            'description': 'Pointer to the last message. It can be used as a '
            '``start``/``end`` for another ``get_messages`` query.',
            'example': 't2-10202_143892_626_3663_5949_6_558_7501_0',
        },
    )

    start_time = MillisecondsTimestamp(
        required=True,
        nullable=True,
        metadata={
            'description': 'The oldest timestamp of the returned messages',
            'example': '2022-07-23T16:20:01.254223',
        },
    )

    end_time = MillisecondsTimestamp(
        required=True,
        nullable=True,
        metadata={
            'description': 'The newest timestamp of the returned messages',
            'example': '2022-07-23T18:20:01.254223',
        },
    )



from dataclasses import dataclass, field
from socket import AddressFamily
from typing import List, Optional

from marshmallow.validate import OneOf


@dataclass
class NetworkInterface:
    """
    Network interface statistics data class.
    """

    interface: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Network interface identifier',
                'example': 'eth0',
            }
        }
    )

    bytes_sent: int = field(
        metadata={
            'metadata': {
                'description': 'Number of bytes sent',
            }
        }
    )

    bytes_recv: int = field(
        metadata={
            'metadata': {
                'description': 'Number of bytes received',
            }
        }
    )

    packets_sent: int = field(
        metadata={
            'metadata': {
                'description': 'Number of packets sent',
            }
        }
    )

    packets_recv: int = field(
        metadata={
            'metadata': {
                'description': 'Number of packets received',
            }
        }
    )

    errors_in: int = field(
        metadata={
            'metadata': {
                'description': 'Number of errors on incoming traffic',
            },
        }
    )

    errors_out: int = field(
        metadata={
            'metadata': {
                'description': 'Number of errors on outgoing traffic',
            },
        }
    )

    drop_in: int = field(
        metadata={
            'metadata': {
                'description': 'Number of packets dropped on incoming traffic',
            },
        }
    )

    drop_out: int = field(
        metadata={
            'metadata': {
                'description': 'Number of packets dropped on outgoing traffic',
            },
        }
    )

    is_up: bool = field(
        metadata={
            'metadata': {
                'description': 'Whether the interface is active',
                'example': True,
            },
        }
    )

    speed: int = field(
        metadata={
            'metadata': {
                'description': 'Interface reported speed in Mbps',
                'example': 10000,
            },
        }
    )

    mtu: int = field(
        metadata={
            'metadata': {
                'description': 'Interface maximum transmission unit expressed '
                'in bytes',
                'example': 65535,
            },
        }
    )

    duplex: str = field(
        metadata={
            'validate': OneOf(['FULL', 'HALF', 'UNKNOWN']),
            'metadata': {
                'description': 'Interface duplex configuration. Can be FULL, '
                'HALF or UNKNOWN',
                'example': 'FULL',
            },
        }
    )

    flags: List[str] = field(
        default_factory=list,
        metadata={
            'metadata': {
                'description': 'List of flags associated to the interface',
                'example': ['up', 'broadcast', 'running'],
            }
        },
    )

    addresses: List['NetworkInterfaceAddress'] = field(
        default_factory=list,
        metadata={
            'metadata': {
                'description': 'List of addresses associated to the interface',
                'example': [
                    {
                        'family': AddressFamily.AF_INET.name,
                        'address': '192.168.1.2',
                        'netmask': '255.255.255.0',
                        'broadcast': '192.168.1.255',
                    }
                ],
            },
        },
    )


@dataclass
class NetworkInterfaceAddress:
    """
    Network interface address data class.
    """

    family: AddressFamily = field(
        metadata={
            'metadata': {
                'description': 'Address family',
                'example': AddressFamily.AF_INET.name,
            }
        }
    )

    address: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'IPv4 or IPv6 address of the interface',
                'example': '192.168.1.2',
            }
        }
    )

    netmask: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Netmask for the interface address',
                'example': '255.255.255.0',
            }
        }
    )

    broadcast: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Broadcast address for the interface',
                'example': '192.168.1.255',
            }
        }
    )



from abc import ABC
from typing import IO, Iterable, List, Optional, Type, Union

import numpy as np
import sounddevice as sd

from platypush.message.event.sound import (
    SoundPlaybackPausedEvent,
    SoundPlaybackResumedEvent,
    SoundPlaybackStartedEvent,
    SoundPlaybackStoppedEvent,
)

from ..._converters import RawOutputAudioConverter
from ..._model import StreamType
from .._base import AudioThread


class AudioPlayer(AudioThread, ABC):
    """
    Base ``AudioPlayer`` class.

    An ``AudioPlayer`` thread is responsible for playing audio (either from a
    file/URL or from a synthetic source) to an output device, writing it to the
    converter process and dispatching the converted audio to the registered
    consumers.
    """

    def __init__(
        self, *args, sound: Optional[Union[dict, Iterable[dict]]] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sound = sound

    @classmethod
    def build(
        cls,
        infile: Optional[str] = None,
        sound: Optional[Union[dict, Iterable[dict]]] = None,
        **kwargs,
    ) -> "AudioPlayer":
        from ._resource import AudioResourcePlayer
        from ._synth import AudioSynthPlayer, Sound

        if infile:
            return AudioResourcePlayer(infile=infile, **kwargs)
        if sound:
            sounds: List[dict] = (  # type: ignore
                [sound] if isinstance(sound, dict) else sound
            )

            return AudioSynthPlayer(sounds=[Sound.build(**s) for s in sounds], **kwargs)

        raise AssertionError('Either infile or url must be specified')

    @property
    def direction(self) -> StreamType:
        return StreamType.OUTPUT

    def _on_converter_timeout(self, *_, **__) -> bool:
        return False  # break

    @property
    def _stream_type(self) -> Type[sd.RawOutputStream]:
        return sd.RawOutputStream

    @property
    def _audio_converter_type(self) -> Type[RawOutputAudioConverter]:
        return RawOutputAudioConverter

    def _on_audio_converted(self, data: bytes, out_f: Optional[IO] = None):
        if self.audio_stream:
            self.audio_stream.write(
                np.asarray(
                    self.gain
                    * np.frombuffer(data, dtype=self.dtype).reshape(-1, self.channels),
                    dtype=self.dtype,
                )
            )

        super()._on_audio_converted(data, out_f)

    @property
    def _started_event_type(self) -> Type[SoundPlaybackStartedEvent]:
        return SoundPlaybackStartedEvent

    @property
    def _stopped_event_type(self) -> Type[SoundPlaybackStoppedEvent]:
        return SoundPlaybackStoppedEvent

    @property
    def _paused_event_type(self) -> Type[SoundPlaybackPausedEvent]:
        return SoundPlaybackPausedEvent

    @property
    def _resumed_event_type(self) -> Type[SoundPlaybackResumedEvent]:
        return SoundPlaybackResumedEvent


# vim:sw=4:ts=4:et:



from collections.abc import Iterable
from multiprocessing import Process
from typing import Dict, List, Optional
from time import sleep

import hid

from platypush.context import get_bus
from platypush.message.event.hid import (
    HidDeviceConnectedEvent,
    HidDeviceDataEvent,
    HidDeviceDisconnectedEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.hid import HidDeviceSchema, HidMonitoredDeviceSchema


class HidPlugin(RunnablePlugin):
    """
    This plugin can be used to interact directly with HID devices (including
    Bluetooth, USB and several serial and wireless devices) over the raw
    interface.

    This is the preferred way of communicating with joypads.

    Note that on Linux reading from the devices requires the user running the
    Platypush service to have (at least) read access to the ``/dev/hidraw*``
    devices. However, it is still possible to get connected/disconnected events
    even without having to open a connection to the device. A way to make HID
    raw devices accessible to e.g. a particular user group is via udev rules.
    For example, create a file named ``/etc/udev/rules.d/99-hid.rules`` with
    the following content::

        # Make all /dev/hidraw* devices accessible in read/write to users in
        # the group input
        KERNEL=="hidraw*", GROUP="input", MODE="0660"

    A more granular solution is to provide read and/or write access only to a
    specific device that you want to access, and only to the specific user
    running the Platypush service::

        KERNEL=="hidraw*", ATTRS{idVendor}=="1234", ATTRS{idProduct}=="5678", USER="user", MODE="0660"

    If you don't want to reboot the device after adding the rules, then you can
    reload the rules for udev service and re-trigger them::

        # udevadm control --reload && udevadm trigger

    """

    def __init__(
        self,
        monitored_devices: Optional[Iterable[dict]] = None,
        poll_seconds: int = 1,
        **kwargs,
    ):
        """
        :param monitored_devices: Map of devices that should be monitored for
            new data. Format (note that all the device filtering attributes are
            optional):

            .. schema:: hid.HidMonitoredDeviceSchema(many=True)

        :param poll_seconds: How often the plugin should check for changes in
            the list of devices (default: 1 second).
        """
        super().__init__(**kwargs)
        self._poll_seconds = poll_seconds
        self._filters = HidMonitoredDeviceSchema().load(
            monitored_devices or [], many=True
        )

        self._device_monitors: Dict[str, Process] = {}
        self._available_devices: Dict[str, dict] = {}

    def main(self):
        while not self.should_stop():
            scanned_devices = {
                dev['path']: dev for dev in self.get_devices().output  # type: ignore
            }

            self._handle_device_events(scanned_devices)
            self._available_devices = scanned_devices
            if self._poll_seconds and self._poll_seconds > 0:
                sleep(self._poll_seconds)

    def stop(self):
        device_monitors = self._device_monitors.copy()
        for monitored_device in device_monitors:
            self._unregister_device_monitor(monitored_device)

        super().stop()

    @action
    def get_devices(self) -> List[dict]:
        """
        Get the HID devices available on the host.

        :return: .. schema:: hid.HidDeviceSchema(many=True)
        """
        return list(  # type: ignore
            {
                dev['path']: HidDeviceSchema().load(dev)
                for dev in hid.enumerate()  # type: ignore
            }.values()
        )

    def _get_monitor_rule(self, device: dict) -> Optional[dict]:
        """
        :return: .. schema:: hid.HidMonitoredDeviceSchema
        """
        matching_rules = [
            rule
            for rule in (self._filters or [])
            if all(
                rule[attr] == device.get(attr)
                for attr in (
                    'path',
                    'serial_number',
                    'vendor_id',
                    'product_id',
                    'manufacturer_string',
                    'product_string',
                )
                if rule.get(attr)
            )
        ]

        if matching_rules:
            return matching_rules[0]

    def _device_monitor(self, dev_def: dict, rule: dict):
        path = dev_def['path']
        data_size = rule['data_size']
        poll_seconds = rule['poll_seconds']
        notify_only_if_changed = rule['notify_only_if_changed']
        last_data = None
        self.logger.info(f'Starting monitor for device {path}')

        def wait():
            if poll_seconds and poll_seconds > 0:
                sleep(poll_seconds)

        while True:
            device = None

            try:
                if not device:
                    device = hid.Device(dev_def['vendor_id'], dev_def['product_id'])  # type: ignore
                data = device.read(data_size)
            except Exception as e:
                if device:
                    self.logger.warning(f'Read error from {path}: {e}')
                device = None
                sleep(5)
                continue

            if not notify_only_if_changed or data != last_data:
                data_dump = ''.join(f'{x:02x}' for x in data)
                get_bus().post(HidDeviceDataEvent(data=data_dump, **dev_def))
                last_data = data

            wait()

    def _register_device_monitor(self, device: dict, rule: dict):
        """
        Register a monitor for a device.
        """
        path = device['path']
        # Make sure that no other monitors are registered for this device
        self._unregister_device_monitor(path)
        monitor_proc = self._device_monitors[path] = Process(
            target=self._device_monitor, args=(device, rule)
        )

        monitor_proc.start()

    def _unregister_device_monitor(self, path: str):
        """
        Unregister a monitor for a device.
        """
        monitor_proc = self._device_monitors.get(path)
        if monitor_proc and monitor_proc.is_alive():
            self.logger.info(f'Terminating monitor for device {path}')

            try:
                monitor_proc.terminate()
                monitor_proc.join(2)
                if monitor_proc.is_alive():
                    monitor_proc.kill()
            except Exception as e:
                self.logger.warning(f'Error while terminating monitor for {path}: {e}')

            del self._device_monitors[path]

    def _handle_new_devices(self, scanned_devices: dict):
        """
        Handles connection events and monitor registering.
        """
        scanned_device_paths = set(scanned_devices.keys())
        available_device_paths = set(self._available_devices)
        new_device_paths = scanned_device_paths.difference(available_device_paths)

        for path in new_device_paths:
            device = scanned_devices[path]
            get_bus().post(HidDeviceConnectedEvent(**device))
            monitor_rule = self._get_monitor_rule(device)
            if monitor_rule:
                self._register_device_monitor(device, monitor_rule)

    def _handle_disconnected_devices(self, scanned_devices: dict):
        """
        Handles disconnection events and monitor unregistering.
        """
        scanned_device_paths = set(scanned_devices.keys())
        available_device_paths = set(self._available_devices)
        lost_device_paths = available_device_paths.difference(scanned_device_paths)

        for path in lost_device_paths:
            device = self._available_devices.get(path)
            if device:
                get_bus().post(HidDeviceDisconnectedEvent(**device))
                self._unregister_device_monitor(device['path'])

    def _handle_device_events(self, scanned_devices: dict):
        """
        Handles connected/disconnected device events and register/unregister the
        monitors based on the user-provided rules when required.
        """
        self._handle_new_devices(scanned_devices)
        self._handle_disconnected_devices(scanned_devices)



import contextlib
import ipaddress
from typing import Collection, Dict, List, Mapping, Optional, Union

from platypush.entities import Entity, SwitchEntityManager
from platypush.entities.switches import Switch
from platypush.plugins import RunnablePlugin, action
from platypush.utils.workers import Workers

from .lib import WemoRunner
from .scanner import Scanner


class SwitchWemoPlugin(RunnablePlugin, SwitchEntityManager):
    """
    Plugin to control a Belkin WeMo smart switches
    (https://www.belkin.com/us/Products/home-automation/c/wemo-home-automation/)
    """

    _default_port = 49153

    def __init__(
        self,
        devices: Optional[Union[Collection[str], Mapping[str, str]]] = None,
        netmask: Optional[str] = None,
        port: int = _default_port,
        **kwargs
    ):
        """
        This plugin previously used ``ouimeaux`` for auto-discovery, but it's
        been dropped because:

            1. Too slow
            2. Too heavy
            3. Auto-discovery failed too often

        However, this also means that you now have to specify either:

            - ``devices``: The devices you want to control, as a static list/map
            - ``netmask``: The IP netmask that should be scanned for WeMo devices

        :param devices: List of IP addresses or name->address map containing
            the WeMo Switch devices to control.
        :type devices: list or dict

        :param netmask: Alternatively to a list of static IP->name pairs,  you
            can specify the network mask where the devices should be scanned
            (e.g. '192.168.1.0/24')

        :param port: Port where the WeMo devices are expected to expose the
            RPC/XML over HTTP service (default: 49153)
        """

        super().__init__(**kwargs)
        assert devices or netmask, (
            'Please specify either a static list of devices (either a list of '
            'IP addresses or a name->address map) or an IP netmask to scan for '
            'devices'
        )

        self.port = port
        self.netmask = netmask
        self._devices: Dict[str, str] = {}
        self._init_devices(devices)

    def _init_devices(self, devices):
        if devices:
            self._devices.update(
                devices
                if isinstance(devices, dict)
                else {addr: addr for addr in devices}
            )
        else:
            self._devices = {}

        self._addresses = set(self._devices.values())

    def _get_address(self, device: str) -> str:
        if device not in self._addresses:
            with contextlib.suppress(KeyError):
                return self._devices[device]

        return device

    @action
    # pylint: disable=arguments-differ
    def status(
        self,
        device: Optional[Union[str, Collection[str]]] = None,
        publish_entities: bool = True,
        **__
    ) -> List[dict]:
        if device:
            if isinstance(device, str):
                devices = {device: device}
            else:
                devices = {d: d for d in device}
        else:
            devices = self._devices.copy()

        ret = [
            {
                "id": addr,
                "ip": addr,
                "name": name if name != addr else WemoRunner.get_name(addr),
                "on": WemoRunner.get_state(addr),
            }
            for (name, addr) in devices.items()
        ]

        if publish_entities:
            self.publish_entities(ret)
        return ret

    def transform_entities(self, entities: Collection[dict]) -> List[Entity]:
        return [
            Switch(
                id=dev["id"],
                name=dev["name"],
                state=dev["on"],
                data={
                    "ip": dev["ip"],
                },
            )
            for dev in (entities or [])
        ]

    @action
    def on(self, device: str, **_):  # pylint: disable=arguments-differ
        """
        Turn a switch on

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.on(device)
        return self.status(device)

    @action
    def off(self, device: str, **_):  # pylint: disable=arguments-differ
        """
        Turn a switch off

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.off(device)
        return self.status(device)

    @action
    def toggle(self, device: str, *_, **__):  # pylint: disable=arguments-differ
        """
        Toggle a device on/off state

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.toggle(device)
        return self.status(device)

    @action
    def get_state(self, device: str):
        """
        Get the on state of a device (True/False)

        :param device: Device name or address
        """
        device = self._get_address(device)
        return WemoRunner.get_state(device)

    @action
    def get_name(self, device: str):
        """
        Get the friendly name of a device

        :param device: Device name or address
        """
        device = self._get_address(device)
        return WemoRunner.get_name(device)

    @action
    def scan(
        self, netmask: Optional[str] = None, publish_entities: bool = True
    ) -> List[dict]:
        netmask = netmask or self.netmask
        assert netmask, "Scan not supported: No netmask specified"

        workers = Workers(10, Scanner, port=self.port)
        with workers:
            for addr in ipaddress.IPv4Network(netmask):
                workers.put(addr.exploded)

        devices = {dev.name: dev.addr for dev in workers.responses}

        self._init_devices(devices)
        return self.status(publish_entities=publish_entities).output

    def main(self):
        def scan():
            status = (
                self.scan(publish_entities=False).output
                if not self._devices
                else self.status(self._devices.values(), publish_entities=False).output
            )

            return {dev['ip']: dev for dev in status}

        devices = {}

        while not self.should_stop():
            new_devices = scan()
            updated_devices = {
                ip: new_devices[ip]
                for ip, dev in new_devices.items()
                if any(v != devices.get(ip, {}).get(k) for k, v in dev.items())
            }

            if updated_devices:
                self.publish_entities(updated_devices.values())

            devices = new_devices
            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



import os
from typing import Iterable, Optional, Type, Union

import aioxmpp
import aioxmpp.im

from platypush.config import Config
from platypush.message.event.xmpp import XmppConnectedEvent
from platypush.plugins import AsyncRunnablePlugin, action

from ._base import XmppBasePlugin
from ._config import XmppConfig
from ._handlers import (
    XmppBaseHandler,
    XmppConnectionHandler,
    XmppConversationHandler,
    XmppHandlersRegistry,
    XmppPresenceHandler,
    XmppRoomHandler,
    XmppRosterHandler,
    discover_handlers,
)
from ._mixins import XmppBaseMixin
from ._state import SerializedState, StateSerializer
from ._types import Errors, XmppPresence


# pylint: disable=too-many-ancestors
class XmppPlugin(AsyncRunnablePlugin, XmppBasePlugin):
    """
    XMPP integration.
    """

    def __init__(
        self,
        user_id: str,
        password: Optional[str] = None,
        language: Optional[str] = None,
        anonymous: bool = False,
        auto_accept_invites: bool = True,
        restore_state: bool = True,
        state_file: Optional[str] = None,
        **kwargs,
    ):
        """
        :param user_id: Jabber/user ID, in the format ``user@example.org``.
        :param password: User password.
        :param language: ISO string for the language code that will be used by
            the bot (default: ``None``).
        :param anonymous: Whether to use anonymous authentication (default:
            ``False``).
        :param auto_accept_invites: Whether to automatically accept invites to
            conversations (default: True). If set to False, and you still want
            some control on which invites should be accepted, you can create a
            ``hook`` on
            :class:`platypush.message.event.xmpp.XmppRoomInviteEvent` that
            calls either :meth:`.accept_invite` or :meth:`.reject_invite` with
            the ``room_id`` specified on the event, if it is a room event, or
            subscribe to
            :class:`platypush.message.event.xmpp.XmppContactAddRequestEvent`
            and call either :meth:`.accept_invite` or :meth:`.reject_invite`
            with the ``user_id`` specified on the event, if it is a contact add
            request.
        :param restore_state: If ``True`` (default) then any previously joined
            conversation or subscribed contact will be joined/subscribed again
            when the plugin restarts. Otherwise, upon restart the plugin will
            start from a state with no subscriptions nor joined rooms.
        :param state_file: Path where the previous state will be stored, if
            ``restore_state`` is ``True``. Default:
            ``<WORKDIR>/xmpp/state.json``.
        """
        super(XmppBasePlugin, self).__init__(user_id=user_id, language=language)
        super(AsyncRunnablePlugin, self).__init__(**kwargs)

        self._security = aioxmpp.make_security_layer(password, anonymous=anonymous)
        self._config = XmppConfig(
            auto_accept_invites=auto_accept_invites,
            restore_state=restore_state,
            state_file=os.path.expanduser(
                state_file or os.path.join(Config.get_workdir(), 'xmpp', 'state.json')
            ),
        )
        self._loaded_state = SerializedState()
        self._state_serializer = StateSerializer(user_id=self._jid, config=self._config)
        self._handlers = XmppHandlersRegistry(self)
        self.restore_state()

    def restore_state(self):
        """
        Reload the previous state from the configured state file.
        """
        if not (self._config.state_file and self._config.restore_state):
            return

        self._loaded_state = self._state_serializer.load()

    @property
    def _conn_handler(self) -> XmppConnectionHandler:
        return self._handlers[XmppConnectionHandler]

    @property
    def _conv_handler(self) -> XmppConversationHandler:
        return self._handlers[XmppConversationHandler]

    @property
    def _presence_handler(self) -> XmppPresenceHandler:
        return self._handlers[XmppPresenceHandler]

    @property
    def _room_handler(self) -> XmppRoomHandler:
        return self._handlers[XmppRoomHandler]

    @property
    def _roster_handler(self) -> XmppRosterHandler:
        return self._handlers[XmppRosterHandler]

    def _on_disconnect(self, reason: Optional[Union[str, Exception]] = None):
        self._conn_handler.disconnect(reason)

    def _register_handlers(self):
        for hndl_type in discover_handlers():
            hndl = self.register_xmpp_handler(hndl_type)
            hndl.restore_state()

    def register_xmpp_handler(self, hndl_type: Type[XmppBaseMixin]) -> XmppBaseHandler:
        self.logger.debug('Registering handler: %s', hndl_type)
        self._handlers[hndl_type] = hndl_type(
            user_id=self._jid,
            language=self._lang,
            config=self._config,
            state=self._state,
            client=self._client,
            loop=self._loop,
            state_serializer=self._state_serializer,
            loaded_state=self._loaded_state,
        )

        return self._handlers[hndl_type]

    def should_stop(self) -> bool:
        return super().should_stop() or self._state.should_stop.is_set()

    def stop(self):
        self._state.should_stop.set()
        self._stop_state_serializer()
        self._stop_client()
        self._on_disconnect(reason='Plugin terminated')
        super().stop()

    def _stop_state_serializer(self):
        if self._state_serializer:
            self._state_serializer.flush()
            self._state_serializer.wait(self._state_serializer.flush_timeout)

    def _stop_client(self):
        if self._client:
            self._client.stop()
            self._client = None

    async def listen(self):
        self._client = aioxmpp.PresenceManagedClient(self._jid, self._security)

        try:
            async with self._client.connected():
                self._register_handlers()
                self._post_event(XmppConnectedEvent)
                await self._state.should_stop.wait()
        except Exception as e:
            self.logger.warning('XMPP connection error: %s', e)
            self.logger.exception(e)
            self._on_disconnect(e)
            raise e

    @action
    def send_message(
        self,
        body: str,
        user_id: Optional[str] = None,
        room_id: Optional[str] = None,
        language: Optional[str] = None,
    ):
        """
        Send a message to a target (the Jabber ID of another user or room).

        :param body: Message body.
        :param user_id: Jabber ID of the target user. Either user_id or room_id
            should be specified.
        :param room_id: Jabber ID of the target room. Either user_id or room_id
            should be specified.
        :param language: Override the default language code.
        """
        if room_id:
            self._room_handler.send_message(
                room_id=room_id, body=body, language=language
            )
        elif user_id:
            self._conv_handler.send_message(
                user_id=user_id, body=body, language=language
            )
        else:
            raise AssertionError(Errors.USER_ID_OR_ROOM_ID)

    @action
    def join(
        self,
        room_id: str,
        nick: Optional[str] = None,
        password: Optional[str] = None,
        auto_rejoin: bool = True,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Join a room/conversation.

        :param room_id: The Jabber ID of the conversation to join.
        :param nick: The nickname that the bot should use in the room (default:
            the nickname specified in the configuration's ``user_id``
            parameter).
        :param password: The password of the room (default: None).
        :param auto_rejoin: Whether to automatically rejoin the room after
            disconnection/kick (default: True).
        :param timeout: Room join timeout (default: 20 seconds). Set to null
            for no timeout.
        """
        nick = nick or self._jid.localpart
        self._async_run(
            self._room_handler.join,
            room_id,
            timeout=timeout,
            nick=nick,
            password=password,
            auto_rejoin=auto_rejoin,
        )

    @action
    def leave(
        self, room_id: str, timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT
    ):
        """
        Leave a room/conversation.

        :param room_id: The Jabber ID of the conversation to leave.
        :param timeout: Room leave timeout (default: 20 seconds). Set to null
            for no timeout.
        """
        self._async_run(
            self._room_handler.leave,
            room_id,
            timeout=timeout,
        )

    @action
    def accept_invite(
        self, room_id: Optional[str] = None, user_id: Optional[str] = None
    ):
        """
        Accept a pending invite to a multi-user conversation or a contact add
        request.

        :param user_id: The target ``user_id`` if this is a contact add request.
        :param room_id: The target ``room_id`` if this is a room invite request.
        """
        if room_id:
            self._room_handler.accept_invite(room_id)
        elif user_id:
            self._roster_handler.accept_invite(user_id)
        else:
            raise AssertionError(Errors.USER_ID_OR_ROOM_ID)

    @action
    def reject_invite(
        self, room_id: Optional[str] = None, user_id: Optional[str] = None
    ):
        """
        Reject a pending invite to a multi-user conversation or a contact add
        request.

        :param user_id: The target ``user_id`` if this is a contact add request.
        :param room_id: The target ``room_id`` if this is a room invite request.
        """
        if room_id:
            self._room_handler.reject_invite(room_id)
        elif user_id:
            self._roster_handler.reject_invite(user_id)
        else:
            raise AssertionError(Errors.USER_ID_OR_ROOM_ID)

    @action
    def invite(
        self,
        room_id: str,
        user_id: str,
        mode: str = 'direct',
        text: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Invite a user to a room.

        :param room_id: The target room JID.
        :param user_id: The JID of the user to invite.
        :param timeout: Invite request send timeout (default: 20 seconds). Set
            to null for no timeout.
        :param mode: Invite mode - can be ``direct`` (default) or ``mediated``.

           - ``direct``: The invitation is sent directly to the invitee,
             without going through a service specific to the conversation.

           - ``mediated``: The invitation is sent indirectly through a service
             which is providing the conversation. Advantages of using this mode
             include most notably that the service can automatically add the
             invitee to the list of allowed participants in configurations
             where such restrictions exist (or deny the request if the inviter
             does not have the permissions to do so).
        :param text: Optional text to send with the invitation.
        """
        self._async_run(
            self._room_handler.invite,
            room_id=room_id,
            user_id=aioxmpp.JID.fromstr(user_id),
            mode=getattr(
                aioxmpp.im.InviteMode, mode.upper(), aioxmpp.im.InviteMode.DIRECT
            ),
            text=text,
            timeout=timeout,
        )

    @action
    def set_presence(self, presence: Union[str, XmppPresence]):
        """
        Set/broadcast a new presence state for the user.

        :param presence: The new presence state. Possible values are:

            - ``available``
            - ``offline``
            - ``away``
            - ``xa``
            - ``chat``
            - ``dnd``

        """
        pres = XmppPresence(presence.lower()) if isinstance(presence, str) else presence
        self._presence_handler.set_presence(pres)

    @action
    def set_affiliation(
        self,
        room_id: str,
        user_id: str,
        affiliation: str,
        reason: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Change the affiliation of a user to a room.

        :param room_id: The target room JID.
        :param user_id: The user JID.
        :param affiliation: The affiliation to set. Possible values are:

            - ``owner``
            - ``member``
            - ``none``
            - ``outcast``
            - ``publisher``
            - ``publish-only``

        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        :param reason: Optional reason for the change.
        """
        self._async_run(
            self._room_handler.set_affiliation,
            room_id=room_id,
            user_id=aioxmpp.JID.fromstr(user_id),
            affiliation=affiliation,
            reason=reason,
            timeout=timeout,
        )

    @action
    def set_role(
        self,
        room_id: str,
        user_id: str,
        role: str,
        reason: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Change the role of a user in a room.

        :param room_id: The target room JID.
        :param user_id: The user JID.
        :param role: The role to set. Possible values are:

            - ``none``
            - ``participant``
            - ``visitor``
            - ``moderator``

        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        :param reason: Optional reason for the change.
        """
        self._async_run(
            self._room_handler.set_role,
            room_id=room_id,
            user_id=aioxmpp.JID.fromstr(user_id),
            role=role,
            reason=reason,
            timeout=timeout,
        )

    @action
    def kick(
        self,
        room_id: str,
        user_id: str,
        reason: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Kick a user from a room.

        :param room_id: The target room JID.
        :param user_id: The JID of the user to kick.
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        :param reason: Kick reason.
        """
        self._async_run(
            self._room_handler.kick,
            room_id=room_id,
            user_id=aioxmpp.JID.fromstr(user_id),
            reason=reason,
            timeout=timeout,
        )

    @action
    def ban(
        self,
        room_id: str,
        user_id: str,
        reason: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Ban a user from a room.

        :param room_id: The target room JID.
        :param user_id: The JID of the user to ban.
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        :param reason: Ban reason.
        """
        self._async_run(
            self._room_handler.ban,
            room_id=room_id,
            user_id=aioxmpp.JID.fromstr(user_id),
            reason=reason,
            timeout=timeout,
        )

    @action
    def set_topic(
        self,
        room_id: str,
        topic: str,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Set the topic of a room.

        :param room_id: The target room JID.
        :param topic: New topic.
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        """
        self._async_run(
            self._room_handler.set_topic,
            room_id=room_id,
            topic=topic,
            timeout=timeout,
        )

    @action
    def set_room_configuration(
        self,
        room_id: str,
        name: Optional[bool] = None,
        description: Optional[bool] = None,
        members_only: Optional[bool] = None,
        persistent: Optional[bool] = None,
        moderated: Optional[bool] = None,
        allow_invites: Optional[bool] = None,
        allow_private_messages: Optional[bool] = None,
        allow_change_subject: Optional[bool] = None,
        enable_logging: Optional[bool] = None,
        max_history_fetch: Optional[int] = None,
        max_users: Optional[int] = None,
        password_protected: Optional[bool] = None,
        public: Optional[bool] = None,
        room_admins: Optional[Iterable[str]] = None,
        room_owners: Optional[Iterable[str]] = None,
        password: Optional[str] = None,
        language: Optional[str] = None,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Changes the configuration of a room.

        All the parameters are optional, and only those that have a non-null
        value will be set.

        :param room_id: The target room JID.
        :param name: New room name.
        :param description: New room description.
        :param members_only: Whether or not this room is only for members.
        :param persistent: Whether or not this room is persistent.
        :param moderated: Whether or not this room is moderated.
        :param allow_invites: Whether or not this room allows invites.
        :param allow_private_messages: Whether or not this room allows private
            messages.
        :param allow_change_subject: Whether or not this room allows changing
            its subject.
        :param enable_logging: Whether or not this room has logging enabled.
        :param max_history_fetch: Maximum number of past messages to fetch when
            joining the room.
        :param max_users: Maximum number of users allowed in the room.
        :param password_protected: Whether or not this room is password protected.
        :param public: Whether or not this room is publicly visible.
        :param room_admins: List of room admins, by Jabber ID.
        :param room_owners: List of room owners, by Jabber ID.
        :param password: If the room is password protected, configure its
            password here.
        :param language: Language of the room (ISO 2-letter code).
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        """
        self._async_run(
            self._room_handler.set_room_config,
            room_id=room_id,
            name=name,
            description=description,
            members_only=members_only,
            persistent=persistent,
            moderated=moderated,
            allow_invites=allow_invites,
            allow_private_messages=allow_private_messages,
            allow_change_subject=allow_change_subject,
            enable_logging=enable_logging,
            max_history_fetch=max_history_fetch,
            max_users=max_users,
            password_protected=password_protected,
            public=public,
            room_admins=room_admins,
            room_owners=room_owners,
            password=password,
            language=language,
            timeout=timeout,
        )

    @action
    def set_nick(
        self,
        room_id: str,
        nick: str,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
    ):
        """
        Set the nick of the user on a specific room.

        :param room_id: The target room JID.
        :param nick: New nick.
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        """
        self._async_run(
            self._room_handler.set_nick,
            room_id=room_id,
            nick=nick,
            timeout=timeout,
        )

    @action
    def add_user(self, user_id: str):
        """
        Add the specified user ID to the roster.

        :param user_id: The Jabber ID of the user to add.
        """
        self._roster_handler.add_user(user_id)

    @action
    def remove_user(self, user_id: str):
        """
        Remove the specified user ID from the roster.

        :param user_id: The Jabber ID of the user to remove.
        """
        self._roster_handler.remove_user(user_id)

    @action
    def request_voice(
        self, room_id: str, timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT
    ):
        """
        Request voice (i.e. participant role) in a room.

        :param room_id: The Jabber ID of the room.
        :param timeout: Request timeout (default: 20 seconds). Set to null for
            no timeout.
        """
        self._async_run(
            self._room_handler.request_voice, room_id=room_id, timeout=timeout
        )

    @action
    def status(self):
        """
        Get the current status of the client.

        :return:

            .. code-block:: python

                {
                  # List of pending room invites, as Jabber IDs
                  "room_invites": ["bar@conference.xmpp.example.org"],

                  # List of pending user invites, as Jabber IDs
                  "user_invites": ["ignore-me@example.org"],

                  # List of users the client is subscribed to
                  "users": [
                    "buddy@example.org"
                  ],

                  # Map of rooms the client has joined, indexed by room ID
                  "rooms": {
                    "tests@conference.xmpp.manganiello.tech": {
                      "room_id": "foo@conference.xmpp.example.org",
                      "joined": true,
                      # Possible values:
                      # ACTIVE, DISCONNECTED, HISTORY, JOIN_PRESENCE
                      "state": "ACTIVE",
                      "nick": "me",

                      # Map of room members, indexed by user ID
                      "members": {
                        "me@example.org": {
                          "user_id": "me@example.org",
                          "nick": "me",
                          # Possible affiliation values:
                          # none, member, outcast, owner, publisher, publish-only
                          "affiliation": "none",

                          # Possible role values:
                          # none, participant, visitor, moderator
                          "role": "participant",
                          "is_self": true,
                          "available": true,

                          # Possible state values:
                          # available, offline, away, xa, chat, dnd
                          "state": "available"
                        },

                        "buddy@example.org": {
                          "user_id": "buddy@example.org",
                          "nick": "SomeBuddy",
                          "affiliation": "owner",
                          "role": "moderator",
                          "is_self": false,
                          "available": true,
                          "state": "away"
                        }
                      }
                    }
                  }
                }

        """
        return self._state.asdict(return_passwords=False)


# vim:sw=4:ts=4:et:



import aioxmpp

from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors,too-few-public-methods
class XmppPubSubHandler(XmppBaseHandler):
    """
    Handler for XMPP pub/sub events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pubsub: aioxmpp.PubSubClient = self._client.summon(aioxmpp.PubSubClient)



import enum
from typing import Optional

from platypush.message.event import Event


class ZeroconfEventType(enum.Enum):
    ADD = 'add'
    UPDATE = 'update'
    REMOVE = 'remove'


class ZeroconfEvent(Event):
    def __init__(self, service_event: ZeroconfEventType, service_type: str, service_name: str,
                 service_info: Optional[dict] = None, *args, **kwargs):
        super().__init__(*args, service_event=service_event.value, service_type=service_type,
                         service_name=service_name, service_info=service_info, **kwargs)

        self.service_type = service_type
        self.service_name = service_name
        self.service_info = service_info


class ZeroconfServiceAddedEvent(ZeroconfEvent):
    """
    Event triggered when a service is added or discovered.
    """
    def __init__(self, *args, **kwargs):
        kwargs['service_event'] = ZeroconfEventType.ADD
        super().__init__(*args, **kwargs)


class ZeroconfServiceUpdatedEvent(ZeroconfEvent):
    """
    Event triggered when a service is updated.
    """
    def __init__(self, *args, **kwargs):
        kwargs['service_event'] = ZeroconfEventType.UPDATE
        super().__init__(*args, **kwargs)


class ZeroconfServiceRemovedEvent(ZeroconfEvent):
    """
    Event triggered when a service is removed.
    """
    def __init__(self, *args, **kwargs):
        kwargs['service_event'] = ZeroconfEventType.REMOVE
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:



import os
import subprocess
import threading
from typing import Callable, Optional, List

from platypush.plugins import Plugin, action


class FfmpegPlugin(Plugin):
    """
    Generic FFmpeg plugin to interact with media files and devices.
    """

    def __init__(
        self, ffmpeg_cmd: str = 'ffmpeg', ffprobe_cmd: str = 'ffprobe', **kwargs
    ):
        super().__init__(**kwargs)
        self.ffmpeg_cmd = ffmpeg_cmd
        self.ffprobe_cmd = ffprobe_cmd
        self._threads = {}
        self._next_thread_id = 1
        self._thread_lock = threading.RLock()

    @action
    def info(self, filename: str, **kwargs) -> dict:
        """
        Get the information of a media file.

        :param filename: Path to the media file.
        :return: Media file information. Example:

          .. code-block:: json

              {
                "streams": [
                    {
                        "index": 0,
                        "codec_name": "h264",
                        "codec_long_name": "H.264 / AVC / MPEG-4 AVC / MPEG-4 part 10",
                        "profile": "High 4:2:2",
                        "codec_type": "video",
                        "codec_time_base": "1/60",
                        "codec_tag_string": "[0][0][0][0]",
                        "codec_tag": "0x0000",
                        "width": 640,
                        "height": 480,
                        "coded_width": 640,
                        "coded_height": 480,
                        "closed_captions": 0,
                        "has_b_frames": 2,
                        "pix_fmt": "yuv422p",
                        "level": 30,
                        "chroma_location": "left",
                        "field_order": "progressive",
                        "refs": 1,
                        "is_avc": "true",
                        "nal_length_size": "4",
                        "r_frame_rate": "30/1",
                        "avg_frame_rate": "30/1",
                        "time_base": "1/1000",
                        "start_pts": 0,
                        "start_time": "0.000000",
                        "bits_per_raw_sample": "8",
                        "disposition": {
                            "default": 1,
                            "dub": 0,
                            "original": 0,
                            "comment": 0,
                            "lyrics": 0,
                            "karaoke": 0,
                            "forced": 0,
                            "hearing_impaired": 0,
                            "visual_impaired": 0,
                            "clean_effects": 0,
                            "attached_pic": 0,
                            "timed_thumbnails": 0
                        },
                        "tags": {
                            "ENCODER": "Lavc58.91.100 libx264"
                        }
                    }
                ],
                "format": {
                    "filename": "./output.mkv",
                    "nb_streams": 1,
                    "nb_programs": 0,
                    "format_name": "matroska,webm",
                    "format_long_name": "Matroska / WebM",
                    "start_time": "0.000000",
                    "size": "786432",
                    "probe_score": 100,
                    "tags": {
                        "ENCODER": "Lavf58.45.100"
                    }
                }
            }

        """
        # noinspection PyPackageRequirements
        import ffmpeg

        filename = os.path.abspath(os.path.expanduser(filename))
        info = ffmpeg.probe(filename, cmd=self.ffprobe_cmd, **kwargs)
        return info

    @staticmethod
    def _poll_thread(
        proc: subprocess.Popen,
        packet_size: int,
        on_packet: Callable[[bytes], None],
        on_open: Optional[Callable[[], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
    ):
        try:
            if on_open:
                on_open()

            while proc.poll() is None:
                data = proc.stdout.read(packet_size)
                on_packet(data)
        finally:
            if on_close:
                on_close()

    @action
    def start(
        self,
        pipeline: List[dict],
        pipe_stdin: bool = False,
        pipe_stdout: bool = False,
        pipe_stderr: bool = False,
        quiet: bool = False,
        overwrite_output: bool = False,
        on_packet: Callable[[bytes], None] = None,
        packet_size: int = 4096,
    ):
        # noinspection PyPackageRequirements
        import ffmpeg

        stream = ffmpeg

        for step in pipeline:
            args = step.pop('args') if 'args' in step else []
            stream = getattr(stream, step.pop('method'))(*args, **step)

        self.logger.info(
            'Executing {cmd} {args}'.format(cmd=self.ffmpeg_cmd, args=stream.get_args())
        )
        proc = stream.run_async(
            cmd=self.ffmpeg_cmd,
            pipe_stdin=pipe_stdin,
            pipe_stdout=pipe_stdout,
            pipe_stderr=pipe_stderr,
            quiet=quiet,
            overwrite_output=overwrite_output,
        )

        if on_packet:
            with self._thread_lock:
                self._threads[self._next_thread_id] = threading.Thread(
                    target=self._poll_thread,
                    kwargs={
                        'proc': proc,
                        'on_packet': on_packet,
                        'packet_size': packet_size,
                    },
                )

                self._threads[self._next_thread_id].start()
                self._next_thread_id += 1


# vim:sw=4:ts=4:et:



import logging
from typing import Optional

from .. import MediaPlugin


class MediaSearcher:
    """
    Base class for media searchers
    """

    def __init__(self, *args, media_plugin: Optional[MediaPlugin] = None, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.media_plugin = media_plugin

    def search(self, query, *args, **kwargs):
        raise NotImplementedError('The search method should be implemented ' +
                                  'by a derived class')


from .local import LocalMediaSearcher
from .youtube import YoutubeMediaSearcher
from .torrent import TorrentMediaSearcher
from .plex import PlexMediaSearcher
from .jellyfin import JellyfinMediaSearcher

__all__ = [
    'MediaSearcher', 'LocalMediaSearcher', 'TorrentMediaSearcher',
    'YoutubeMediaSearcher', 'PlexMediaSearcher', 'JellyfinMediaSearcher',
]


# vim:sw=4:ts=4:et:



import socket

from dataclasses import dataclass
from typing import Optional

from platypush.utils.workers import Worker
from .lib import WemoRunner


@dataclass
class ScanResult:
    """
    Models a scan result.
    """

    addr: str
    name: str
    on: bool


class Scanner(Worker):
    """
    Worker class used to scan WeMo devices on the network.
    """

    timeout = 1.5

    def __init__(self, *args, port: int = WemoRunner.default_port, **kwargs):
        super().__init__(*args, **kwargs)
        self.port = port

    def process(self, msg: str) -> Optional[ScanResult]:
        addr = msg

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.timeout)
            sock.connect((addr, self.port))
            sock.close()

            return ScanResult(
                addr=addr, name=WemoRunner.get_name(addr), on=WemoRunner.get_state(addr)
            )
        except OSError:
            return None


# vim:sw=4:ts=4:et:



from threading import Timer
from multiprocessing import Process
from typing import Iterable, Optional

import Leap

from platypush.context import get_bus
from platypush.message.event.sensor.leap import (
    LeapFrameEvent,
    LeapFrameStartEvent,
    LeapFrameStopEvent,
    LeapConnectEvent,
    LeapDisconnectEvent,
)
from platypush.plugins import RunnablePlugin


class LeapPlugin(RunnablePlugin):
    """
    Integration to handle events from a `Leap Motion
    <https://www.leapmotion.com/>`_ device to track hands and
    gestures.

    Note that the default SDK is not compatible with Python 3. Follow the
    instructions on https://github.com/BlackLight/leap-sdk-python3 to build the
    Python 3 module.

    Also, you'll need the Leap driver and utils installed on your OS (follow
    instructions at https://www.leapmotion.com/setup/) and the `leapd` daemon
    running to recognize your controller.

    Requires:

        * The Leap Motion SDK compiled with Python 3 support, see my port at
            https://github.com:BlackLight/leap-sdk-python3.git
        * The ``leapd`` daemon to be running and your Leap Motion connected

    """

    _listener_proc = None

    def __init__(
        self,
        position_ranges: Optional[Iterable[Iterable[float]]] = None,
        position_tolerance: float = 0.0,
        poll_interval: Optional[float] = 0.1,
        **kwargs
    ):
        """
        :param position_ranges: It specifies how wide the hand space (x, y and
            z axes) should be in millimiters.

            Default::

                [
                    [-300.0, 300.0],  # x axis
                    [25.0, 600.0],    # y axis
                    [-300.0, 300.0],  # z axis
                ]
        :param position_tolerance: % of change between a frame and the next to
            really consider the next frame as a new one (default: 0).
        :param poll_interval: How often the plugin should generate and push
            events (default: at most one event each 0.1 seconds). Note that a
            Leap Motion generates events with a very high throughput, so you
            may want to set a higher value if you want to throttle the events
            and avoid flooding the bus.
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        if position_ranges is None:
            position_ranges = [
                [-300.0, 300.0],  # x axis
                [25.0, 600.0],  # y axis
                [-300.0, 300.0],  # z axis
            ]

        self.position_ranges = position_ranges
        self.position_tolerance = position_tolerance

    def main(self):
        def _listener_process():
            listener = LeapListener(
                position_ranges=self.position_ranges,
                position_tolerance=self.position_tolerance,
                frames_throttle_secs=self.poll_interval,
                logger=self.logger,
            )

            controller = Leap.Controller()

            if not controller:
                raise RuntimeError(
                    'No Leap Motion controller found - is your '
                    + 'device connected and is leapd running?'
                )

            controller.add_listener(listener)
            self.logger.info('Leap Motion device initialized')

            try:
                self.wait_stop()
            except KeyboardInterrupt:
                self.logger.info('Terminating Leap Motion listener')

        while not self.should_stop():
            self.wait_stop(1)
            self._listener_proc = Process(target=_listener_process)
            self._listener_proc.start()
            self._listener_proc.join()
            self._listener_proc = None


class LeapFuture(Timer):
    def __init__(self, seconds, listener, event):
        self.listener = listener
        self.event = event
        super().__init__(seconds, self._callback_wrapper())

    def _callback_wrapper(self):
        def _callback():
            self.listener._send_event(self.event)  # pylint: disable=protected-access

        return _callback


class LeapListener(Leap.Listener):
    def __init__(
        self, position_ranges, position_tolerance, logger, frames_throttle_secs=None
    ):
        super().__init__()

        self.prev_frame = None
        self.position_ranges = position_ranges
        self.position_tolerance = position_tolerance
        self.frames_throttle_secs = frames_throttle_secs
        self.logger = logger
        self.running_future = None

    def _send_event(self, event):
        get_bus().post(event)

    def send_event(self, event):
        if self.frames_throttle_secs:
            if not self.running_future or not self.running_future.is_alive():
                self.running_future = LeapFuture(
                    seconds=self.frames_throttle_secs, listener=self, event=event
                )
                self.running_future.start()
        else:
            self._send_event(event)

    def on_init(self, *_, **__):
        self.prev_frame = None
        self.logger.info('Leap controller listener initialized')

    def on_connect(self, *_, **__):
        self.logger.info('Leap controller connected')
        self.prev_frame = None
        self.send_event(LeapConnectEvent())

    def on_disconnect(self, *_, **__):
        self.logger.info('Leap controller disconnected')
        self.prev_frame = None
        self.send_event(LeapDisconnectEvent())

    def on_exit(self, *_, **__):
        self.logger.info('Leap listener terminated')

    def on_frame(self, controller):
        frame = controller.frame()

        if len(frame.hands) > 0:
            hands = self._flatten_hands(frame)
            if hands:
                if not self.prev_frame:
                    self.send_event(LeapFrameStartEvent())
                self.send_event(LeapFrameEvent(hands=hands))
            self.prev_frame = frame
        else:
            if self.prev_frame:
                self.send_event(LeapFrameStopEvent())
            self.prev_frame = None

    def _flatten_hands(self, frame):
        return [
            {
                'confidence': hand.confidence,
                'direction': [hand.direction[0], hand.direction[1], hand.direction[2]],
                'id': hand.id,
                'is_left': hand.is_left,
                'is_right': hand.is_right,
                'palm_normal': [
                    hand.palm_normal[0],
                    hand.palm_normal[1],
                    hand.palm_normal[2],
                ],
                'palm_position': self._normalize_position(hand.palm_position),
                'palm_velocity': [
                    hand.palm_velocity[0],
                    hand.palm_velocity[1],
                    hand.palm_velocity[2],
                ],
                'palm_width': hand.palm_width,
                'sphere_center': [
                    hand.sphere_center[0],
                    hand.sphere_center[1],
                    hand.sphere_center[2],
                ],
                'sphere_radius': hand.sphere_radius,
                'stabilized_palm_position': self._normalize_position(
                    hand.stabilized_palm_position
                ),
                'time_visible': hand.time_visible,
                'wrist_position': self._normalize_position(hand.wrist_position),
            }
            for i, hand in enumerate(frame.hands)
            if hand.is_valid
            and (
                len(frame.hands) != len(self.prev_frame.hands)
                or self._position_changed(
                    old_position=self.prev_frame.hands[i].stabilized_palm_position,
                    new_position=hand.stabilized_palm_position,
                )
                if self.prev_frame
                else True
            )
        ]

    def _normalize_position(self, position):
        # Normalize absolute position onto a hemisphere centered in (0,0)
        # having x_range = z_range = [-100, 100], y_range = [0, 100]

        return [
            self._scale_scalar(
                value=position[0],
                range=self.position_ranges[0],
                new_range=[-100.0, 100.0],
            ),
            self._scale_scalar(
                value=position[1], range=self.position_ranges[1], new_range=[0.0, 100.0]
            ),
            self._scale_scalar(
                value=position[2],
                range=self.position_ranges[2],
                new_range=[-100.0, 100.0],
            ),
        ]

    @staticmethod
    def _scale_scalar(value, range, new_range):
        if value < range[0]:
            value = range[0]
        if value > range[1]:
            value = range[1]

        return ((new_range[1] - new_range[0]) / (range[1] - range[0])) * (
            value - range[0]
        ) + new_range[0]

    def _position_changed(self, old_position, new_position):
        return (
            abs(old_position[0] - new_position[0]) > self.position_tolerance
            or abs(old_position[1] - new_position[1]) > self.position_tolerance
            or abs(old_position[2] - new_position[2]) > self.position_tolerance
        )


# vim:sw=4:ts=4:et:



from queue import Queue, Empty
from threading import Event
from time import time
from typing import List, Optional

from platypush.entities import Entity


class EntitiesQueue(Queue):
    """
    Extends the ``Queue`` class to provide an abstraction that allows to
    getting and putting multiple entities at once and synchronize with the
    upstream caller.
    """

    def __init__(self, stop_event: Optional[Event] = None, timeout: float = 1.0):
        super().__init__()
        self._timeout = timeout
        self._should_stop = stop_event

    @property
    def should_stop(self) -> bool:
        return self._should_stop.is_set() if self._should_stop else False

    def get(self, block=True, timeout=None) -> List[Entity]:
        """
        Returns a batch of entities read from the queue.
        """
        timeout = timeout or self._timeout
        entities = []
        last_poll_time = time()

        while not self.should_stop and (time() - last_poll_time < timeout):
            try:
                entity = super().get(block=block, timeout=0.5)
            except Empty:
                continue

            if entity:
                entities.append(entity)

        return entities

    def put(self, *entities: Entity, block=True, timeout=None):
        """
        This method is called by an entity manager to update and persist the
        state of some entities.
        """
        for entity in entities:
            super().put(entity, block=block, timeout=timeout)



import datetime
import enum
import os
import time
import threading
from random import randint
from typing import Callable, Optional, Union

import croniter
from dateutil.parser import isoparse
from dateutil.tz import gettz

from platypush.context import get_bus, get_plugin
from platypush.entities.alarm import Alarm as AlarmDb
from platypush.message.request import Request
from platypush.message.event.alarm import (
    AlarmDisabledEvent,
    AlarmDismissedEvent,
    AlarmEnabledEvent,
    AlarmSnoozedEvent,
    AlarmStartedEvent,
)
from platypush.plugins.media import MediaPlugin, PlayerState
from platypush.procedure import Procedure


class AlarmState(enum.IntEnum):
    """
    Alarm states.
    """

    WAITING = 1
    RUNNING = 2
    DISMISSED = 3
    SNOOZED = 4
    SHUTDOWN = 5
    UNKNOWN = -1


class AlarmConditionType(enum.Enum):
    """
    Alarm condition types.
    """

    CRON = 'cron'
    INTERVAL = 'interval'
    TIMESTAMP = 'timestamp'


class Alarm:
    """
    Alarm model and controller.
    """

    def __init__(
        self,
        when: Union[str, int, float],
        actions: Optional[Union[list, Procedure]] = None,
        name: Optional[str] = None,
        media: Optional[str] = None,
        media_plugin: Optional[str] = None,
        media_repeat: bool = True,
        audio_volume: Optional[Union[int, float]] = None,
        snooze_interval: float = 300,
        dismiss_interval: float = 300,
        poll_interval: float = 2,
        enabled: bool = True,
        static: bool = False,
        stop_event: Optional[threading.Event] = None,
        on_change: Optional[Callable[['Alarm'], None]] = None,
        **_,
    ):
        self.id = randint(0, 65535)
        self.when = when
        self.name = name or f'Alarm_{self.id}'
        self.media = self._get_media_resource(media)
        self.media_plugin = media_plugin
        self.media_repeat = media_repeat
        self.audio_volume = audio_volume
        self.snooze_interval = snooze_interval
        self.dismiss_interval = dismiss_interval
        self.state = AlarmState.UNKNOWN
        self.timer: Optional[threading.Timer] = None
        self.static = static
        self.actions = (
            actions
            if isinstance(actions, Procedure)
            else Procedure.build(
                name=name, _async=False, requests=actions or [], id=self.id
            )
        )

        self._enabled = enabled
        self._runtime_snooze_interval = snooze_interval
        self.stop_event = stop_event or threading.Event()
        self.poll_interval = poll_interval
        self.on_change = on_change
        self._dismiss_timer: Optional[threading.Timer] = None

    def _on_change(self):
        if self.on_change:
            self.on_change(self)

    @staticmethod
    def _get_media_resource(media: Optional[str]) -> Optional[str]:
        if not media:
            return None

        if media.startswith('file://'):
            media = media[len('file://') :]

        media_path = os.path.abspath(os.path.expanduser(media))
        if os.path.isfile(media_path):
            media = media_path

        return media

    @property
    def is_cron(self) -> bool:
        if not isinstance(self.when, str):
            return False

        try:
            croniter.croniter(self.when, time.time()).get_next()
            return True
        except (AttributeError, croniter.CroniterBadCronError):
            return False

    @property
    def is_interval(self) -> bool:
        try:
            float(self.when)
            return True
        except (TypeError, ValueError):
            return False

    @property
    def is_timestamp(self) -> bool:
        if not isinstance(self.when, str):
            return False

        try:
            datetime.datetime.fromisoformat(self.when)
            return True
        except Exception:
            return False

    @property
    def condition_type(self) -> AlarmConditionType:
        if self.is_cron:
            return AlarmConditionType.CRON
        if self.is_interval:
            return AlarmConditionType.INTERVAL
        if self.is_timestamp:
            return AlarmConditionType.TIMESTAMP

        raise ValueError(f'Invalid alarm condition {self.when}')

    def get_next(self) -> Optional[float]:
        now = time.time()
        t = 0

        try:
            # If when is a number, interpret it as number of seconds into the future
            delta = float(self.when)
            t = now + delta
            self.when = datetime.datetime.fromtimestamp(t).isoformat()
        except (TypeError, ValueError):
            assert isinstance(self.when, str), f'Invalid alarm time {self.when}'

            try:
                # If when is a cron expression, get the next run time
                t = croniter.croniter(
                    self.when,
                    datetime.datetime.fromtimestamp(now).replace(tzinfo=gettz()),
                ).get_next()
            except (AttributeError, croniter.CroniterBadCronError):
                try:
                    # If when is an ISO-8601 timestamp, parse it
                    t = isoparse(self.when).timestamp()
                except Exception as e:
                    raise AssertionError(f'Invalid alarm time {self.when}: {e}') from e

        return t if t >= now else None

    def is_enabled(self):
        return self._enabled

    def is_shut_down(self):
        return self.state in {
            AlarmState.SHUTDOWN,
            AlarmState.DISMISSED,
            AlarmState.UNKNOWN,
        }

    def is_expired(self):
        return (self.get_next() or 0) < time.time()

    def disable(self):
        self.set_enabled(False)

    def enable(self):
        self.set_enabled(True)

    def set_enabled(self, enabled: bool):
        if enabled == self._enabled:
            return

        self._enabled = enabled
        evt_type = AlarmEnabledEvent if enabled else AlarmDisabledEvent
        get_bus().post(evt_type(name=self.name))
        self._on_change()

    def dismiss(self):
        self.state = AlarmState.DISMISSED
        self.stop_audio()
        self._clear_dismiss_timer()
        get_bus().post(AlarmDismissedEvent(name=self.name))
        self._on_change()

    def snooze(self, interval: Optional[float] = None):
        self._runtime_snooze_interval = interval or self.snooze_interval
        self.state = AlarmState.SNOOZED
        self.stop_audio()
        self._clear_dismiss_timer()
        get_bus().post(
            AlarmSnoozedEvent(name=self.name, interval=self._runtime_snooze_interval)
        )
        self._on_change()

    def start(self):
        if self.timer:
            self.timer.cancel()

        next_run = self.get_next()
        if next_run is None:
            return

        interval = next_run - time.time()
        self.state = AlarmState.WAITING
        self.timer = threading.Timer(interval, self.alarm_callback)
        self.timer.start()
        self._clear_dismiss_timer()
        self._on_change()

    def stop(self):
        self.state = AlarmState.SHUTDOWN
        self.stop_audio()

        if self.timer:
            self.timer.cancel()
            self.timer = None

        self._on_change()

    def _clear_dismiss_timer(self):
        if self._dismiss_timer:
            self._dismiss_timer.cancel()
            self._dismiss_timer = None

    def _get_media_plugin(self) -> MediaPlugin:
        plugin = get_plugin(self.media_plugin)
        assert plugin and isinstance(plugin, MediaPlugin), (
            f'Invalid audio plugin {self.media_plugin}'
            if plugin
            else f'Missing audio plugin {self.media_plugin}'
        )

        return plugin

    def play_audio(self):
        def thread():
            self._get_media_plugin().play(self.media)
            if self.audio_volume is not None:
                self._get_media_plugin().set_volume(self.audio_volume)

        audio_thread = threading.Thread(target=thread)
        audio_thread.start()

    def stop_audio(self):
        self._get_media_plugin().stop()

    def _on_start(self):
        if self.state != AlarmState.RUNNING:
            self._dismiss_timer = threading.Timer(self.dismiss_interval, self.dismiss)
            self._dismiss_timer.start()

        self.state = AlarmState.RUNNING
        get_bus().post(AlarmStartedEvent(name=self.name))
        self._on_change()
        if self.media_plugin and self.media:
            self.play_audio()

        self.actions.execute()

    def _on_running(self):
        sleep_time = None

        while not self.should_stop():
            plugin_status = self._get_media_plugin().status().output
            if not isinstance(plugin_status, dict):
                self.wait_stop(self.poll_interval)
                continue

            state = plugin_status.get('state')
            if state == PlayerState.STOP.value:
                if self.state == AlarmState.SNOOZED:
                    sleep_time = self._runtime_snooze_interval
                else:
                    if (
                        self.media_repeat
                        and self.state != AlarmState.DISMISSED
                        and not self.should_stop()
                    ):
                        self.wait_stop(self.poll_interval)
                        if not self.should_stop():
                            self.play_audio()
                            continue

                    self.state = AlarmState.WAITING

                break

            self._on_change()
            self.wait_stop(self.poll_interval)

        return sleep_time

    def alarm_callback(self):
        while not self.should_stop():
            if self.is_enabled():
                self._on_start()
            elif self.state != AlarmState.WAITING:
                self.state = AlarmState.WAITING
                self._on_change()

            self.wait_stop(self.poll_interval)
            sleep_time = None
            if self.state == AlarmState.RUNNING:
                sleep_time = self._on_running()

            if self.state == AlarmState.SNOOZED:
                sleep_time = self._runtime_snooze_interval
            elif self.get_next() is None:
                self.state = AlarmState.SHUTDOWN
                break

            if not sleep_time:
                next_run = self.get_next()
                sleep_time = (
                    next_run - time.time()
                    if next_run is not None
                    else self.poll_interval
                )

            self.stop_event.wait(sleep_time)

    def wait_stop(self, timeout: Optional[float] = None):
        self.stop_event.wait(timeout)

    def should_stop(self):
        return (
            self.stop_event.is_set()
            or (self.is_expired() and self.state == AlarmState.DISMISSED)
            or self.state == AlarmState.SHUTDOWN
        )

    def to_dict(self) -> dict:
        return {
            'id': self.name,
            'type': 'alarm',
            'name': self.name,
            'when': self.when,
            'next_run': self.get_next(),
            'enabled': self.is_enabled(),
            'state': self.state.name,
            'media': self.media,
            'media_plugin': self.media_plugin,
            'media_repeat': self.media_repeat,
            'audio_volume': self.audio_volume,
            'snooze_interval': self.snooze_interval,
            'dismiss_interval': self.dismiss_interval,
            'actions': self.actions.requests,
            'static': self.static,
            'condition_type': self.condition_type.value,
        }

    @classmethod
    def from_db(cls, alarm: AlarmDb, **kwargs) -> 'Alarm':
        return cls(
            when=str(alarm.when),
            name=str(alarm.name),
            media=alarm.media,  # type: ignore
            media_plugin=kwargs.pop('media_plugin', alarm.media_plugin),  # type: ignore
            media_repeat=alarm.media_repeat,  # type: ignore
            audio_volume=alarm.audio_volume,  # type: ignore
            actions=alarm.actions,  # type: ignore
            snooze_interval=alarm.snooze_interval,  # type: ignore
            dismiss_interval=alarm.dismiss_interval,  # type: ignore
            enabled=bool(alarm.enabled),
            static=bool(alarm.static),
            state=getattr(AlarmState, str(alarm.state)),
            **kwargs,
        )

    def to_db(self) -> AlarmDb:
        return AlarmDb(
            id=self.name,
            name=self.name,
            when=self.when,
            state=self.state.name,
            next_run=self.get_next(),
            media=self.media,
            media_plugin=self.media_plugin,
            media_repeat=self.media_repeat,
            audio_volume=self.audio_volume,
            actions=[
                Request.to_dict(req) if isinstance(req, Request) else req
                for req in self.actions.requests
            ],
            snooze_interval=self.snooze_interval,
            dismiss_interval=self.dismiss_interval,
            enabled=self.is_enabled(),
            static=self.static,
            condition_type=self.condition_type.value,
        )


# vim:sw=4:ts=4:et:



from datetime import datetime, date
from enum import Enum
from typing import Any, Optional, Type, Union

from dateutil.parser import isoparse
from dateutil.tz import tzutc
from marshmallow import fields


class StrippedString(fields.Function):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        kwargs['serialize'] = self._strip
        kwargs['deserialize'] = self._strip
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        if getattr(obj, attr, None) is not None:
            return self._strip(getattr(obj, attr))
        if getattr(obj, 'get', lambda _: None)(attr) is not None:
            return self._strip(obj.get(attr))

    @staticmethod
    def _strip(value: str):
        return value.strip()


class Function(fields.Function):  # lgtm [py/missing-call-to-init]
    def _get_attr(self, obj, attr: str, _recursive=True):
        if self.attribute and _recursive:
            return self._get_attr(obj, self.attribute, False)
        if hasattr(obj, attr):
            return getattr(obj, attr)
        elif hasattr(obj, 'get'):
            return obj.get(attr)


class DateTime(Function):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.metadata = {
            'example': datetime.now(tz=tzutc()).isoformat(),
            **(self.metadata or {}),
        }

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        value = normalize_datetime(self._get_attr(obj, attr))
        if value:
            return value.isoformat()

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[datetime]:
        return normalize_datetime(value)


class Date(Function):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.metadata = {
            'example': date.today().isoformat(),
            **(self.metadata or {}),
        }

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        value = normalize_datetime(self._get_attr(obj, attr))
        if value:
            return date(value.year, value.month, value.day).isoformat()

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[date]:
        dt = normalize_datetime(value)
        return date.fromtimestamp(dt.timestamp())


class EnumField(Function):
    """
    Field that maps enum values.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, *args, type: Type[Enum], **kwargs):
        super().__init__(*args, **kwargs)
        self.type = type

    def _serialize(self, value: Optional[Enum], *_, **__) -> Optional[Any]:
        return value.value if value is not None else None

    def _deserialize(self, value: Optional[Any], *_, **__) -> Optional[Any]:
        return self.type(value) if value is not None else None


def normalize_datetime(
    dt: Optional[Union[str, date, datetime]]
) -> Optional[Union[date, datetime]]:
    if not dt:
        return
    if isinstance(dt, (datetime, date)):
        return dt

    try:
        dt = float(dt)
        return datetime.fromtimestamp(dt)
    except (TypeError, ValueError):
        return isoparse(dt)



from dataclasses import dataclass
from typing import Final


@dataclass
class Errors:
    """
    A static class to model plugin error messages.
    """

    CLIENT: Final[str] = 'The XMPP client is not connected'
    HANDLERS: Final[str] = 'No registered XMPP handlers found'
    LOOP: Final[str] = 'The event loop is not running'
    NO_INVITE: Final[str] = 'No such conversation invite'
    NO_USER: Final[str] = 'No such user'
    ROOM_NOT_JOINED: Final[str] = 'The bot has not joined this room'
    USER_ID_OR_ROOM_ID: Final[str] = 'You should specify either user_id or room_id'



from typing import Optional

from pychromecast import (
    CastBrowser,
    Chromecast,
    ChromecastConnectionError,
    SimpleCastListener,
    get_chromecast_from_cast_info,
)

from platypush.backend.http.app.utils import get_remote_base_url
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.media import MediaPlugin
from platypush.utils import get_mime_type
from platypush.message.event.media import MediaPlayRequestEvent

from ._listener import MediaListener
from ._subtitles import SubtitlesAsyncHandler
from ._utils import convert_status, post_event


class MediaChromecastPlugin(MediaPlugin, RunnablePlugin):
    """
    Plugin to control Chromecast devices.
    """

    STREAM_TYPE_UNKNOWN = "UNKNOWN"
    STREAM_TYPE_BUFFERED = "BUFFERED"
    STREAM_TYPE_LIVE = "LIVE"

    def __init__(
        self, chromecast: Optional[str] = None, poll_interval: float = 30, **kwargs
    ):
        """
        :param chromecast: Default Chromecast to cast to if no name is specified.
        :param poll_interval: How often the plugin should poll for new/removed
            Chromecast devices (default: 30 seconds).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self._is_local = False
        self.chromecast = chromecast
        self._chromecasts_by_uuid = {}
        self._chromecasts_by_name = {}
        self._media_listeners = {}
        self._zc = None
        self._browser = None

    @property
    def zc(self):
        from zeroconf import Zeroconf

        if not self._zc:
            self._zc = Zeroconf()

        return self._zc

    @property
    def browser(self):
        if not self._browser:
            self._browser = CastBrowser(
                SimpleCastListener(self._on_chromecast_discovered), self.zc
            )

            self._browser.start_discovery()

        return self._browser

    def _on_chromecast_discovered(self, _, service: str):
        self.logger.info('Discovered Chromecast: %s', service)

    @staticmethod
    def _get_device_property(cc, prop: str):
        if hasattr(cc, 'device'):  # Previous pychromecast API
            return getattr(cc.device, prop)
        return getattr(cc.cast_info, prop)

    def _serialize_device(self, cc: Chromecast) -> dict:
        """
        Convert a Chromecast object and its status to a dictionary.
        """
        if hasattr(cc, 'cast_info'):  # Newer PyChromecast API
            host = cc.cast_info.host
            port = cc.cast_info.port
        elif hasattr(cc, 'host'):
            host = getattr(cc, 'host', None)
            port = getattr(cc, 'port', None)
        elif hasattr(cc, 'uri'):
            host, port = cc.uri.split(':')
        else:
            raise RuntimeError('Invalid Chromecast object')

        return {
            'type': cc.cast_type,
            'name': cc.name,
            'manufacturer': self._get_device_property(cc, 'manufacturer'),
            'model_name': cc.model_name,
            'uuid': str(cc.uuid),
            'address': host,
            'port': port,
            'status': (
                {
                    'app': {
                        'id': cc.app_id,
                        'name': cc.app_display_name,
                    },
                    'is_active_input': cc.status.is_active_input,
                    'is_stand_by': cc.status.is_stand_by,
                    'is_idle': cc.is_idle,
                    'namespaces': cc.status.namespaces,
                    'volume': round(100 * cc.status.volume_level, 2),
                    'muted': cc.status.volume_muted,
                    **convert_status(cc.media_controller.status),
                }
                if cc.status
                else {}
            ),
        }

    def _event_callback(self, _, cast: Chromecast):
        self._chromecasts_by_uuid[cast.uuid] = cast
        self._chromecasts_by_name[
            self._get_device_property(cast, 'friendly_name')
        ] = cast

    def get_chromecast(self, chromecast=None):
        if isinstance(chromecast, Chromecast):
            return chromecast

        if self._chromecasts_by_uuid.get(chromecast):
            return self._chromecasts_by_uuid[chromecast]

        if self._chromecasts_by_name.get(chromecast):
            return self._chromecasts_by_name[chromecast]

        raise AssertionError(f'Chromecast {chromecast} not found')

    @action
    def play(
        self,
        resource: str,
        *_,
        content_type: Optional[str] = None,
        chromecast: Optional[str] = None,
        title: Optional[str] = None,
        image_url: Optional[str] = None,
        autoplay: bool = True,
        current_time: int = 0,
        stream_type: str = STREAM_TYPE_BUFFERED,
        subtitles: Optional[str] = None,
        subtitles_lang: str = 'en-US',
        subtitles_mime: str = 'text/vtt',
        subtitle_id: int = 1,
        **__,
    ):
        """
        Cast media to an available Chromecast device.

        :param resource: Media to cast
        :param content_type: Content type as a MIME type string
        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        :param title: Optional title
        :param image_url: URL of the image to use for the thumbnail
        :param autoplay: Set it to false if you don't want the content to start
            playing immediately (default: true)
        :param current_time: Time to start the playback in seconds (default: 0)
        :param stream_type: Type of stream to cast. Can be BUFFERED (default),
            LIVE or UNKNOWN
        :param subtitles: URL of the subtitles to be shown
        :param subtitles_lang: Subtitles language (default: en-US)
        :param subtitles_mime: Subtitles MIME type (default: text/vtt)
        :param subtitle_id: ID of the subtitles to be loaded (default: 1)
        """

        if not chromecast:
            chromecast = self.chromecast

        post_event(MediaPlayRequestEvent, resource=resource, device=chromecast)
        cast = self.get_chromecast(chromecast)
        mc = cast.media_controller
        resource = self._get_resource(resource)

        if not content_type:
            content_type = get_mime_type(resource)

        if not content_type:
            raise RuntimeError(f'content_type required to process media {resource}')

        if not resource.startswith('http://') and not resource.startswith('https://'):
            resource = self._start_streaming(resource)['url']
            resource = get_remote_base_url() + resource
            self.logger.info('HTTP media stream started on %s', resource)

        if self._latest_resource:
            if not title:
                title = self._latest_resource.title
            if not image_url:
                image_url = self._latest_resource.image

        self.logger.info('Playing %s on %s', resource, chromecast)

        mc.play_media(
            resource,
            content_type,
            title=self._latest_resource.title if self._latest_resource else title,
            thumb=image_url,
            current_time=current_time,
            autoplay=autoplay,
            stream_type=stream_type,
            subtitles=subtitles,
            subtitles_lang=subtitles_lang,
            subtitles_mime=subtitles_mime,
            subtitle_id=subtitle_id,
        )

        if subtitles:
            mc.register_status_listener(SubtitlesAsyncHandler(mc, subtitle_id))

        mc.block_until_active()
        if self.volume:
            self.set_volume(volume=self.volume, chromecast=chromecast)

        return self.status(chromecast=chromecast)

    @action
    def load(self, *args, **kwargs):
        """
        Alias for :meth:`.play`.
        """
        return self.play(*args, **kwargs)

    @action
    def pause(self, *_, chromecast: Optional[str] = None, **__):
        """
        Pause the current media on the Chromecast.
        """
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)

        if cast.media_controller.status.player_is_paused:
            cast.media_controller.play()
        elif cast.media_controller.status.player_is_playing:
            cast.media_controller.pause()

        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def stop(self, *_, chromecast: Optional[str] = None, **__):  # type: ignore
        if self.should_stop():
            if self._zc:
                self._zc.close()
                self._zc = None

            if self._browser:
                self._browser.stop_discovery()
                self._browser = None

            return

        chromecast = chromecast or self.chromecast
        if not chromecast:
            return None

        cast = self.get_chromecast(chromecast)
        cast.media_controller.stop()
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def rewind(self, chromecast: Optional[str] = None):
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.media_controller.rewind()
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def set_position(self, position: float, chromecast: Optional[str] = None, **_):
        cast = self.get_chromecast(chromecast or self.chromecast)
        cast.media_controller.seek(position)
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def seek(self, position: float, chromecast: Optional[str] = None, **_):
        return self.forward(chromecast=chromecast, offset=position)

    @action
    def back(self, chromecast: Optional[str] = None, offset: int = 30, **_):
        cast = self.get_chromecast(chromecast or self.chromecast)
        mc = cast.media_controller
        if mc.status.current_time:
            mc.seek(mc.status.current_time - offset)
            cast.wait()

        return self.status(chromecast=chromecast)

    @action
    def forward(self, chromecast: Optional[str] = None, offset: int = 30, **_):
        cast = self.get_chromecast(chromecast or self.chromecast)
        mc = cast.media_controller
        if mc.status.current_time:
            mc.seek(mc.status.current_time + offset)
            cast.wait()

        return self.status(chromecast=chromecast)

    @action
    def is_playing(self, chromecast: Optional[str] = None, **_):
        return self.get_chromecast(
            chromecast or self.chromecast
        ).media_controller.status.player_is_playing

    @action
    def is_paused(self, chromecast: Optional[str] = None, **_):
        return self.get_chromecast(
            chromecast or self.chromecast
        ).media_controller.status.player_is_paused

    @action
    def is_idle(self, chromecast: Optional[str] = None):
        return self.get_chromecast(
            chromecast or self.chromecast
        ).media_controller.status.player_is_idle

    @action
    def list_subtitles(self, chromecast: Optional[str] = None):
        return self.get_chromecast(
            chromecast or self.chromecast
        ).media_controller.status.subtitle_tracks

    @action
    def enable_subtitles(
        self, chromecast: Optional[str] = None, track_id: Optional[int] = None, **_
    ):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        if track_id is not None:
            return mc.enable_subtitle(track_id)
        if mc.status.subtitle_tracks:
            return mc.enable_subtitle(mc.status.subtitle_tracks[0].get('trackId'))

    @action
    def disable_subtitles(
        self, chromecast: Optional[str] = None, track_id: Optional[int] = None, **_
    ):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        if track_id:
            return mc.disable_subtitle(track_id)
        if mc.status.current_subtitle_tracks:
            return mc.disable_subtitle(mc.status.current_subtitle_tracks[0])

    @action
    def toggle_subtitles(self, chromecast: Optional[str] = None, **_):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        all_subs = mc.status.subtitle_tracks
        cur_subs = mc.status.current_subtitle_tracks

        if cur_subs:
            return self.disable_subtitles(chromecast, cur_subs[0])

        return self.enable_subtitles(chromecast, all_subs[0].get('trackId'))

    @action
    def status(self, chromecast: Optional[str] = None):
        """
        :return: The status of a Chromecast (if ``chromecast`` is specified) or
            all the discovered/available Chromecasts. Format:

                .. code-block:: javascript

                    {
                      "type": "cast",  // Can be "cast" or "audio"
                      "name": "Living Room TV",
                      "manufacturer": "Google Inc.",
                      "model_name": "Chromecast",
                      "uuid": "f812afac-80ff-11ee-84dc-001500e8f607",
                      "address": "192.168.1.2",
                      "port": 8009,
                      "status": {
                        "app": {
                          "id": "CC1AD845",
                          "name": "Default Media Receiver"
                        },
                        "is_active_input": false,
                        "is_stand_by": true,
                        "is_idle": true,
                        "namespaces": [
                          "urn:x-cast:com.google.cast.cac",
                          "urn:x-cast:com.google.cast.debugoverlay",
                          "urn:x-cast:com.google.cast.media"
                        ],
                        "volume": 100,
                        "muted": false,
                        "adjusted_current_time": 14.22972,
                        "album_artist": null,
                        "album_name": null,
                        "artist": null,
                        "url": "https://some/video.mp4",
                        "content_type": "video/mp4",
                        "current_subtitle_tracks": [],
                        "position": 1.411891,
                        "duration": 253.376145,
                        "episode": null,
                        "idle_reason": null,
                        "images": [
                          [
                            "https://some/image.jpg",
                            null,
                            null
                          ]
                        ],
                        "last_updated": "2023-11-12T02:03:33.888843",
                        "media_custom_data": {},
                        "media_is_generic": true,
                        "media_is_movie": false,
                        "media_is_musictrack": false,
                        "media_is_photo": false,
                        "media_is_tvshow": false,
                        "media_metadata": {
                          "title": "Some media",
                          "thumb": "https://some/image.jpg",
                          "images": [
                            {
                              "url": "https://some/image.jpg"
                            }
                          ],
                          "metadataType": 0
                        },
                        "media_session_id": 1,
                        "metadata_type": 0,
                        "playback_rate": 1,
                        "player_is_idle": false,
                        "player_is_paused": false,
                        "player_is_playing": true,
                        "state": "play",
                        "season": null,
                        "series_title": null,
                        "stream_type": "BUFFERED",
                        "stream_type_is_buffered": true,
                        "stream_type_is_live": false,
                        "subtitle_tracks": [],
                        "supported_media_commands": 12303,
                        "supports_pause": true,
                        "supports_queue_next": false,
                        "supports_queue_prev": false,
                        "seekable": true,
                        "supports_skip_backward": false,
                        "supports_skip_forward": false,
                        "supports_stream_mute": true,
                        "supports_stream_volume": true,
                        "title": "Some media",
                        "track": null
                      }
                    }

        """
        return self._status(chromecast=chromecast)

    def _status(self, chromecast: Optional[str] = None) -> dict:
        if chromecast:
            assert (
                chromecast in self._chromecasts_by_name
            ), f'No such Chromecast device: {chromecast}'
            return self._serialize_device(self._chromecasts_by_name[chromecast])

        return {
            name: self._serialize_device(cast)
            for name, cast in self._chromecasts_by_name.items()
        }

    @action
    def disconnect(
        self,
        chromecast: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        """
        Disconnect a Chromecast and wait for it to terminate

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        :param timeout: Number of seconds to wait for disconnection (default:
            None: block until termination).
        """
        cast = self.get_chromecast(chromecast)
        cast.disconnect(timeout=timeout)

    @action
    def join(self, chromecast: Optional[str] = None, timeout: Optional[float] = None):
        """
        Blocks the thread until the Chromecast connection is terminated.

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        :param timeout: Number of seconds to wait for disconnection (default:
            None: block until termination).
        """
        cast = self.get_chromecast(chromecast)
        cast.join(timeout=timeout)

    @action
    def quit(self, chromecast: Optional[str] = None):
        """
        Exits the current app on the Chromecast

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        """
        cast = self.get_chromecast(chromecast)
        cast.quit_app()

    @action
    def set_volume(self, volume: float, chromecast: Optional[str] = None):
        """
        Set the Chromecast volume

        :param volume: Volume to be set, between 0 and 100.
        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        """
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.set_volume(volume / 100)
        cast.wait()
        return {
            **self._status(chromecast=chromecast),
            'volume': volume,
        }

    @action
    def volup(self, chromecast: Optional[str] = None, step: float = 10, **_):
        """
        Turn up the Chromecast volume by 10% or step.

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        :param step: Volume increment between 0 and 100 (default: 10%).
        """
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        step /= 100
        cast.volume_up(min(step, 1))
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def voldown(self, chromecast: Optional[str] = None, step: float = 10, **_):
        """
        Turn down the Chromecast volume by 10% or step.

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        :param step: Volume decrement between 0 and 100 (default: 10%).
        """
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        step /= 100
        cast.volume_down(max(step, 0))
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def mute(self, chromecast: Optional[str] = None):
        """
        Toggle the mute status on the Chromecast

        :param chromecast: Chromecast to cast to. If none is specified, then
            the default configured Chromecast will be used.
        """
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.set_volume_muted(not cast.media_controller.status.volume_muted)
        cast.wait()
        return self.status(chromecast=chromecast)

    def set_subtitles(self, *_, **__):
        raise NotImplementedError

    def remove_subtitles(self, *_, **__):
        raise NotImplementedError

    def _refresh_chromecasts(self):
        cast_info = {cast.friendly_name: cast for cast in self.browser.devices.values()}

        for info in cast_info.values():
            name = info.friendly_name
            if self._chromecasts_by_uuid.get(
                info.uuid
            ) and self._chromecasts_by_name.get(name):
                self.logger.debug('Chromecast %s already connected', name)
                continue

            self.logger.info('Started scan for Chromecast %s', name)

            try:
                cc = get_chromecast_from_cast_info(
                    info,
                    self.browser.zc,
                    tries=2,
                    retry_wait=5,
                    timeout=30,
                )

                self._chromecasts_by_name[cc.name] = cc
            except ChromecastConnectionError:
                self.logger.warning('Failed to connect to Chromecast %s', info)
                continue

            if cc.uuid not in self._chromecasts_by_uuid:
                self._chromecasts_by_uuid[cc.uuid] = cc
                self.logger.debug('Connecting to Chromecast %s', name)

                if name not in self._media_listeners:
                    cc.start()
                    self._media_listeners[name] = MediaListener(
                        name=name or str(cc.uuid),
                        cast=cc,
                        callback=self._event_callback,
                    )

                    cc.media_controller.register_status_listener(
                        self._media_listeners[name]
                    )

                    self.logger.info('Connected to Chromecast %s', name)

            self._chromecasts_by_uuid[cc.uuid] = cc
            self._chromecasts_by_name[name] = cc

    def main(self):
        while not self.should_stop():
            try:
                self._refresh_chromecasts()
            finally:
                self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



mock_imports = [
    "Adafruit_IO",
    "Adafruit_Python_DHT",
    "Leap",
    "PIL",
    "PyOBEX",
    "PyOBEX.client",
    "RPLCD",
    "RPi.GPIO",
    "TheengsDecoder",
    "aiofiles",
    "aiofiles.os",
    "aiohttp",
    "aioxmpp",
    "apiclient",
    "async_lru",
    "avs",
    "bcrypt",
    "bleak",
    "bluetooth",
    "bluetooth_numbers",
    "buienradar",
    "cpuinfo",
    "croniter",
    "cups",
    "cv2",
    "cwiid",
    "dbus",
    "deepspeech",
    "defusedxml",
    "dns",
    "docutils",
    "envirophat",
    "feedparser",
    "gevent.wsgi",
    "gi",
    "gi.repository",
    "google",
    "google.api_core",
    "google.auth",
    "google.assistant.embedded",
    "google.assistant.library",
    "google.assistant.library.event",
    "google.assistant.library.file_helpers",
    "google.cloud",
    "google.oauth2.credentials",
    "googlesamples",
    "googlesamples.assistant.grpc.audio_helpers",
    "gps",
    "graphyte",
    "grpc",
    "gunicorn",
    "hid",
    "httplib2",
    "icalendar",
    "imapclient",
    "inotify",
    "inputs",
    "irc",
    "irc.bot",
    "irc.client",
    "irc.connection",
    "irc.events",
    "irc.strings",
    "kafka",
    "keras",
    "linode_api4",
    "luma",
    "marshmallow_dataclass",
    "mpd",
    "ndef",
    "nfc",
    "nio",
    "numpy",
    "oauth2client",
    "oauth2client",
    "omxplayer",
    "openzwave",
    "paho",
    "pandas",
    "paramiko",
    "picamera",
    "plexapi",
    "pmw3901",
    "psutil",
    "pvcheetah",
    "pvleopard",
    "pvorca",
    "pvporcupine",
    "pvrhino",
    "pyHS100",
    "pyaudio",
    "pychromecast",
    "pyclip",
    "pydbus",
    "pydub",
    "pyfirmata2",
    "pyngrok",
    "pyotp",
    "pysmartthings",
    "pyzbar",
    "qrcode",
    "rtmidi",
    "samsungtvws",
    "serial",
    "simple_websocket",
    "smartcard",
    "sounddevice",
    "soundfile",
    "telegram",
    "telegram.ext",
    "tenacity",
    "tensorflow",
    "todoist",
    "trello",
    "twilio",
    "uvicorn",
    "watchdog",
    "wave",
    "websockets",
    "zeroconf",
]
"""
List of modules that should be mocked when building the documentation or running tests.
"""



from datetime import datetime
from typing import Any, Dict, Optional

from platypush.message.event import Event


class MatrixEvent(Event):
    """
    Base matrix event.
    """

    def __init__(
        self,
        *args,
        server_url: str,
        sender_id: Optional[str] = None,
        sender_display_name: Optional[str] = None,
        sender_avatar_url: Optional[str] = None,
        room_id: Optional[str] = None,
        room_name: Optional[str] = None,
        room_topic: Optional[str] = None,
        server_timestamp: Optional[datetime] = None,
        **kwargs
    ):
        """
        :param server_url: Base server URL.
        :param sender_id: The event's sender ID.
        :param sender_display_name: The event's sender display name.
        :param sender_avatar_url: The event's sender avatar URL.
        :param room_id: Event room ID.
        :param room_name: The name of the room associated to the event.
        :param room_topic: The topic of the room associated to the event.
        :param server_timestamp: The server timestamp of the event.
        """
        evt_args: Dict[str, Any] = {
            'server_url': server_url,
        }

        if sender_id:
            evt_args['sender_id'] = sender_id
        if sender_display_name:
            evt_args['sender_display_name'] = sender_display_name
        if sender_avatar_url:
            evt_args['sender_avatar_url'] = sender_avatar_url
        if room_id:
            evt_args['room_id'] = room_id
        if room_name:
            evt_args['room_name'] = room_name
        if room_topic:
            evt_args['room_topic'] = room_topic
        if server_timestamp:
            evt_args['server_timestamp'] = server_timestamp

        super().__init__(*args, **evt_args, **kwargs)


class MatrixSyncEvent(MatrixEvent):
    """
    Event triggered when the startup synchronization has been completed and the
    plugin is ready to use.
    """


class MatrixMessageEvent(MatrixEvent):
    """
    Event triggered when a message is received on a subscribed room.
    """

    def __init__(
        self,
        *args,
        body: str = '',
        url: Optional[str] = None,
        thumbnail_url: Optional[str] = None,
        mimetype: Optional[str] = None,
        formatted_body: Optional[str] = None,
        format: Optional[str] = None,
        **kwargs
    ):
        """
        :param body: The body of the message.
        :param url: The URL of the media file, if the message includes media.
        :param thumbnail_url: The URL of the thumbnail, if the message includes media.
        :param mimetype: The MIME type of the media file, if the message includes media.
        :param formatted_body: The formatted body, if ``format`` is specified.
        :param format: The format of the message (e.g. ``html`` or ``markdown``).
        """
        super().__init__(
            *args,
            body=body,
            url=url,
            thumbnail_url=thumbnail_url,
            mimetype=mimetype,
            formatted_body=formatted_body,
            format=format,
            **kwargs
        )


class MatrixMessageImageEvent(MatrixEvent):
    """
    Event triggered when a message containing an image is received.
    """


class MatrixMessageFileEvent(MatrixEvent):
    """
    Event triggered when a message containing a generic file is received.
    """


class MatrixMessageAudioEvent(MatrixEvent):
    """
    Event triggered when a message containing an audio file is received.
    """


class MatrixMessageVideoEvent(MatrixEvent):
    """
    Event triggered when a message containing a video file is received.
    """


class MatrixReactionEvent(MatrixEvent):
    """
    Event triggered when a user submits a reaction to an event.
    """

    def __init__(self, *args, in_response_to_event_id: str, **kwargs):
        """
        :param in_response_to_event_id: The ID of the URL related to the reaction.
        """
        super().__init__(
            *args, in_response_to_event_id=in_response_to_event_id, **kwargs
        )


class MatrixEncryptedMessageEvent(MatrixMessageEvent):
    """
    Event triggered when a message is received but the client doesn't
    have the E2E keys to decrypt it, or encryption has not been enabled.
    """


class MatrixCallEvent(MatrixEvent):
    """
    Base class for Matrix call events.
    """

    def __init__(
        self, *args, call_id: str, version: int, sdp: Optional[str] = None, **kwargs
    ):
        """
        :param call_id: The unique ID of the call.
        :param version: An increasing integer representing the version of the call.
        :param sdp: SDP text of the session description.
        """
        super().__init__(*args, call_id=call_id, version=version, sdp=sdp, **kwargs)


class MatrixCallInviteEvent(MatrixCallEvent):
    """
    Event triggered when the user is invited to a call.
    """

    def __init__(self, *args, invite_validity: Optional[float] = None, **kwargs):
        """
        :param invite_validity: For how long the invite will be valid, in seconds.
        :param sdp: SDP text of the session description.
        """
        super().__init__(*args, invite_validity=invite_validity, **kwargs)


class MatrixCallAnswerEvent(MatrixCallEvent):
    """
    Event triggered by the callee when they wish to answer the call.
    """


class MatrixCallHangupEvent(MatrixCallEvent):
    """
    Event triggered when a participant in the call exists.
    """


class MatrixRoomCreatedEvent(MatrixEvent):
    """
    Event triggered when a room is created.
    """


class MatrixRoomJoinEvent(MatrixEvent):
    """
    Event triggered when a user joins a room.
    """


class MatrixRoomLeaveEvent(MatrixEvent):
    """
    Event triggered when a user leaves a room.
    """


class MatrixRoomInviteEvent(MatrixEvent):
    """
    Event triggered when the user is invited to a room.
    """


class MatrixRoomTopicChangedEvent(MatrixEvent):
    """
    Event triggered when the topic/title of a room changes.
    """

    def __init__(self, *args, topic: str, **kwargs):
        """
        :param topic: New room topic.
        """
        super().__init__(*args, topic=topic, **kwargs)


class MatrixRoomTypingStartEvent(MatrixEvent):
    """
    Event triggered when a user in a room starts typing.
    """


class MatrixRoomTypingStopEvent(MatrixEvent):
    """
    Event triggered when a user in a room stops typing.
    """


class MatrixRoomSeenReceiptEvent(MatrixEvent):
    """
    Event triggered when the last message seen by a user in a room is updated.
    """


class MatrixUserPresenceEvent(MatrixEvent):
    """
    Event triggered when a user comes online or goes offline.
    """

    def __init__(
        self, *args, is_active: bool, last_active: Optional[datetime], **kwargs
    ):
        """
        :param is_active: True if the user is currently online.
        :param topic: When the user was last active.
        """
        super().__init__(*args, is_active=is_active, last_active=last_active, **kwargs)
